#!/usr/bin/env python
"""
Main window of the iBeatles application
"""

import logging
import os
import warnings

from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QMainWindow

from ibeatles import (
    DEFAULT_NORMALIZATION_ROI,
    DEFAULT_ROI,
    DataType,
    RegionType,
    ScrollBarParameters,
    XAxisMode,
    load_ui,
)
from ibeatles._version import __version__
from ibeatles.about.about_launcher import AboutLauncher
from ibeatles.all_steps.event_handler import EventHandler as GeneralEventHandler
from ibeatles.all_steps.infos_launcher import InfosLauncher
from ibeatles.all_steps.log_launcher import LogHandler, LogLauncher
from ibeatles.all_steps.material import (
    Material,
    MaterialPreDefined,
    MaterialUserDefinedMethod1,
    MaterialUserDefinedMethod2,
)

# import new MVP-based widget
from ibeatles.app.presenters.time_spectra_presenter import TimeSpectraPresenter
from ibeatles.binning.binning_launcher import BinningLauncher
from ibeatles.config_handler import ConfigHandler
from ibeatles.fitting import FittingKeys
from ibeatles.fitting.fitting_launcher import FittingLauncher
from ibeatles.fitting.kropff import KropffThresholdFinder
from ibeatles.session.load_previous_session_launcher import (
    LoadPreviousSessionLauncher,
)
from ibeatles.session.session_handler import SessionHandler
from ibeatles.step1.data_handler import DataHandler
from ibeatles.step1.event_handler import EventHandler as Step1EventHandler
from ibeatles.step1.gui_handler import Step1GuiHandler
from ibeatles.step1.initialization import Initialization
from ibeatles.step1.plot import Step1Plot
from ibeatles.step2.gui_handler import Step2GuiHandler
from ibeatles.step2.initialization import Initialization as Step2Initialization
from ibeatles.step2.normalization import Normalization
from ibeatles.step2.plot import Step2Plot
from ibeatles.step2.reduction_settings_handler import ReductionSettingsHandler
from ibeatles.step2.roi_handler import Step2RoiHandler
from ibeatles.step3.event_handler import EventHandler as Step3EventHandler
from ibeatles.step3.gui_handler import Step3GuiHandler
from ibeatles.step6.strain_mapping_launcher import StrainMappingLauncher
from ibeatles.tools.rotate.rotate_images import RotateImages
from ibeatles.tools.tof_bin.tof_binning_launcher import TofBinningLauncher
from ibeatles.tools.tof_combine.tof_combine_launcher import TofCombineLauncher
from ibeatles.utilities.array_utilities import find_nearest_index
from ibeatles.utilities.bragg_edge_element_handler import BraggEdgeElementHandler
from ibeatles.utilities.bragg_edge_selection_handler import (
    BraggEdgeSelectionHandler,
)
from ibeatles.utilities.get import Get
from ibeatles.utilities.gui_handler import GuiHandler
from ibeatles.utilities.list_data_handler import ListDataHandler
from ibeatles.utilities.retrieve_data_infos import (
    RetrieveGeneralDataInfos,
    RetrieveGeneralFileInfos,
)
from ibeatles.utilities.roi_editor import RoiEditor

warnings.filterwarnings("ignore")


class MainWindow(QMainWindow):
    """ """

    current_tab = 0  # this will be used in case user request to see a tab is not allowed yet
    session_dict = {}  # all the parameters to save to be able to recover the full session

    # log ui
    log_id = None

    config = None  # config such as name of log file, ...

    default_path = {
        DataType.sample: "",
        DataType.ob: "",
        DataType.normalized: "",
        DataType.time_spectra: "",
        DataType.time_spectra_normalized: "",
    }

    list_files = {
        DataType.sample: [],  # ex data_files
        DataType.ob: [],
        DataType.normalized: [],
        DataType.time_spectra: [],
    }

    DEBUGGING = True
    loading_flag = False
    current_data_type = DataType.sample
    cbar = None
    live_data = []
    add_element_editor_ui = None
    roi_editor_ui = {
        DataType.sample: None,
        DataType.ob: None,
        DataType.normalized: None,
    }

    # infos mainwindow
    infos_id = None
    infos_dict = {DataType.sample: None, DataType.ob: None, DataType.normalized: None}
    # ui of infos pushButton
    infos_ui_dict = {
        DataType.sample: None,
        DataType.ob: None,
        DataType.normalized: None,
    }

    # scrollbar below Bragg plot for main 3 data sets
    hkl_scrollbar_ui = {
        "label": {DataType.sample: None, DataType.ob: None, DataType.normalized: None},
        "widget": {DataType.sample: None, DataType.ob: None, DataType.normalized: None},
    }

    hkl_scrollbar_dict = {ScrollBarParameters.maximum: 1, ScrollBarParameters.value: 0}

    # used to report in status bar the error messages
    steps_error = {"step1": {"status": True, "message": ""}}

    # binning window stuff
    binning_ui = None
    binning_line_view = {
        "ui": None,
        "pos": None,
        "adj": None,
        "pen": None,
        "image_view": None,
        "roi": None,
    }

    binning_roi = None  # x0, x1, width, height, bin_size
    # binning_bin_size = 20
    binning_done = False

    # window ui
    tof_combine_ui = None
    tof_rebin_ui = None

    # list hkl, lambda and d0
    list_hkl_lambda_d0_ui = None

    ## FITTING TAB
    # fitting window stuff
    fitting_image_view = None
    there_is_a_roi = False
    init_sigma_alpha_ui = None
    fitting_ui = None
    fitting_story_ui = None
    advanced_selection_ui = None
    fitting_set_variables_ui = None

    kropff_fitting_parameters_viewer_editor_ui = None

    fitting_selection = {"nbr_column": -1, "nbr_row": -1}
    fitting_bragg_edge_linear_selection = []  # [left lambda, right lambda]
    fitting_transparency_slider_value = 50  # from 0 to 100
    display_active_row_flag = True
    # fitting_lr = None
    table_loaded_from_session = False

    # strain mapping ui (step6)
    strain_mapping_ui = None

    # rotate images ui
    rotate_ui = None

    # table dictionary used in fitting/binning/advanced_selection/review
    march_table_dictionary = {}
    table_fitting_story_dictionary = {}
    table_dictionary_from_session = None

    # table dictionary for kropff
    kropff_table_dictionary = {}

    # new entry will be local_bragg_edge_list['new_name'] = {
    #           Material.lattice: value,
    #           Material.crystal_structure: 'FCC',
    #           Material.hkl_d0: None',
    #           Material.method_used: Material.via_lattice_and_crystal_structure}
    local_bragg_edge_list = {}
    selected_element_bragg_edges_array = []
    selected_element_hkl_array = []
    selected_element_name = ""

    # just like above but for user defined ones
    user_defined_bragg_edge_list = {}

    # # [['label', 'x0', 'y0', 'width', 'height', 'group'], ...]
    # init_array = ['label_roi', '0', '0', '20', '20', '0']

    # [[use?, x0, y0, width, height, mean_counts]]
    init_array_normalization = [True, 0, 0, 20, 20, RegionType.background]

    # list roi ui id (when multiple roi in plots)
    list_roi_id = {
        DataType.sample: [],
        DataType.ob: [],
        DataType.normalization: [],
        DataType.normalized: [],
    }

    list_label_roi_id = {
        DataType.sample: [],
        DataType.ob: [],
        DataType.normalization: [],
        DataType.normalized: [],
    }

    list_bragg_edge_selection_id = {
        DataType.sample: None,
        DataType.ob: None,
        DataType.normalized: None,
    }

    list_roi = {
        DataType.sample: [],
        DataType.ob: [],
        DataType.normalization: [],
        DataType.normalized: [],
    }

    old_list_roi = {DataType.sample: [], DataType.ob: [], DataType.normalized: []}

    list_file_selected = {DataType.sample: [], DataType.ob: [], DataType.normalized: []}

    current_bragg_edge_x_axis = {
        DataType.sample: [],
        DataType.ob: [],
        DataType.normalized: [],
        DataType.normalization: [],
    }
    normalized_lambda_bragg_edge_x_axis = []  # will be used by the fitting window

    step2_ui = {
        "area": None,
        "image_view": None,
        "roi": None,
        "bragg_edge_plot": None,
        "normalized_profile_plot": None,
        "caxis": None,
        "xaxis_tof": None,
        "xaxis_lambda": None,
        "xaxis_file_index": None,
        "bragg_edge_selection": None,
    }

    xaxis_button_ui = {
        DataType.sample: {"file_index": None, "tof": None, "lambda": None},
        DataType.ob: {"file_index": None, "tof": None, "lambda": None},
        DataType.normalized: {"file_index": None, "tof": None, "lambda": None},
        DataType.normalization: {"file_index": None, "tof": None, "lambda": None},
    }

    # dictionary that will save the pan and zoom of each of the image view
    image_view_settings = {
        DataType.sample: {
            "state": None,
            "histogram": {
                "mean": None,
                "sum": None,
            },
            "first_time_using_histogram": {
                "mean": True,
                "sum": True,
            },
            "first_time_using_state": False,
        },
        DataType.ob: {
            "state": None,
            "histogram": {
                "mean": None,
                "sum": None,
            },
            "first_time_using_histogram": {
                "mean": True,
                "sum": True,
            },
            "first_time_using_state": False,
        },
        DataType.normalization: {
            "state": None,
            "histogram": None,
            "first_time_using_histogram": True,
            "first_time_using_state": True,
        },
        DataType.normalized: {
            "state": None,
            "histogram": {
                "mean": None,
                "sum": None,
            },
            "first_time_using_histogram": {
                "mean": True,
                "sum": True,
            },
            "first_time_using_state": False,
        },
        DataType.bin: {
            "state": None,
            "histogram": None,
            "first_time_using_histogram": True,
            "first_time_using_state": True,
        },
        DataType.fitting: {
            "state": None,
            "histogram": None,
            "first_time_using_histogram": True,
            "first_time_using_state": True,
        },
    }

    # use to display table that illustrate normalization process in tab2
    normalization_label = {
        "data_ob": "",
        "data": "",
        "no_data": "",
        "previous_status": {"data": False, DataType.ob: False},
    }

    # kropff
    kropff_fitting = None  # pyqtgraph plot
    kropff_is_automatic_bragg_peak_threshold_finder = True
    kropff_automatic_threshold_finder_algorithm = KropffThresholdFinder.sliding_average

    kropff_bragg_peak_good_fit_conditions = {
        "l_hkl_error": {"state": True, "value": 0.01},
        "t_error": {"state": True, "value": 0.01},
        "sigma_error": {"state": True, "value": 0.01},
    }
    kropff_lambda_settings = None

    def __init__(self, parent=None):
        """
        Initialization
        Parameters
        ----------
        """
        # Base class
        super(MainWindow, self).__init__(parent)

        self.ui = load_ui("ui_mainWindow.ui", baseinstance=self)
        self.setWindowTitle("iBeatles")
        self.setup()
        self.init_interface()

        # NOTE: before the complete MVP refactor, we will have to
        #       keep all presenters at the root
        self.time_spectra_presenter = None
        self.normalized_time_spectra_presenter = None

        # list of shutter counts (loaded by tof_binning ui)
        self.shutter_counts = None

        # configuration of config
        o_get = Get(parent=self)
        log_file_name = o_get.get_log_file_name()
        self.log_file_name = log_file_name
        logging.basicConfig(
            filename=log_file_name,
            filemode="w",
            format="[%(levelname)s] - %(asctime)s - %(message)s",
            level=logging.INFO,
        )
        logging.info("*** Starting a new session ***")
        logging.info(f" Version: {__version__}")

        self.automatic_load_of_previous_session()

    def check_log_file_size(self):
        o_handler = LogHandler(parent=self, log_file_name=self.log_file_name)
        o_handler.cut_log_size_if_bigger_than_buffer()

    def init_interface(self):
        o_gui = Initialization(parent=self)
        o_gui.all()
        self.update_delta_lambda()
        self.pre_defined_element_dropBox_changed(0)
        o_material = Material(parent=self)
        o_material.check_status_of_all_widgets()

        # init bragg edge element
        # BraggEdgeElementHandler(parent=self)

        o_gui_2 = Step2Initialization(parent=self)
        o_gui_2.all()

    def setup(self):
        o_config = ConfigHandler(parent=self)
        o_config.load()

        current_folder = None
        if self.config["debugging"]:
            list_homepath = self.config["homepath"]
            for _path in list_homepath:
                if os.path.exists(_path):
                    current_folder = _path
            if current_folder is None:
                current_folder = os.path.expanduser("~")
        else:
            current_folder = os.path.expanduser("~")

        for _key in self.default_path.keys():
            self.default_path[_key] = current_folder

        # self.sample_path = current_folder
        # self.ob_path = current_folder
        # self.normalized_path = current_folder
        # self.time_spectra_path = current_folder
        # self.time_spectra_normalized_path = current_folder

        self.data_metadata = {
            DataType.sample: {
                "title": "Select folder or list of files",
                "list_widget_ui": self.ui.list_sample,
                "folder": current_folder,
                "general_infos": None,
                "data": [],
                FittingKeys.x_axis: "file_index",
                "time_spectra": {
                    "folder": "",
                    "filename": "",
                },
            },
            DataType.ob: {
                "title": "Select folder or list of files",
                "list_widget_ui": self.ui.list_open_beam,
                "folder": current_folder,
                "general_infos": None,
                FittingKeys.x_axis: "file_index",
                "data": [],
            },
            DataType.normalized: {
                "title": "Select folder or list of files",
                "folder": current_folder,
                "general_infos": None,
                "data": [],
                "size": {"width": None, "height": None},
                FittingKeys.x_axis: "file_index",
                "data_live_selection": [],
                "time_spectra": {
                    "folder": "",
                    "filename": "",
                },
            },
            DataType.normalization: {
                "data": [],
            },
            "time_spectra": {
                "title": "Select file",
                "folder": current_folder,
                "normalized_folder": current_folder,
                "general_infos": None,
                "data": [],
                "lambda": [],
                "full_file_name": "",
                "normalized_data": [],
                "normalized_lambda": [],
            },
            DataType.bin: {
                "ui_accessed": False,
            },
            DataType.fitting: {
                "ui_accessed": False,
            },
        }

        self.range_files_to_normalized_step2 = {
            "file_index": [],
            "tof": [],
            "lambda": [],
        }

        self.list_roi[DataType.sample] = [DEFAULT_ROI]
        self.list_roi[DataType.ob] = [DEFAULT_ROI]
        self.list_roi[DataType.normalized] = [DEFAULT_ROI]
        self.list_roi[DataType.normalization] = [DEFAULT_NORMALIZATION_ROI]

        self.old_list_roi[DataType.sample] = [DEFAULT_ROI]
        self.old_list_roi[DataType.ob] = [DEFAULT_ROI]
        self.old_list_roi[DataType.normalized] = [DEFAULT_ROI]
        self.old_list_roi[DataType.normalization] = [DEFAULT_NORMALIZATION_ROI]

        self.infos_ui_dict = {
            DataType.sample: self.ui.sample_infos_pushButton,
            DataType.ob: self.ui.ob_infos_pushButton,
            DataType.normalized: self.ui.normalized_infos_pushButton,
        }

    def automatic_load_of_previous_session(self):
        o_get = Get(parent=self)
        full_config_file_name = o_get.get_automatic_config_file_name()
        if os.path.exists(full_config_file_name):
            load_session_ui = LoadPreviousSessionLauncher(parent=self)
            load_session_ui.show()
        else:
            o_session = SessionHandler(parent=self)
            self.session_dict = o_session.session_dict

    # Menu
    def load_session_clicked(self):
        o_session = SessionHandler(parent=self)
        o_session.load_from_file()
        o_session.load_to_ui()

    def save_session_clicked(self):
        o_session = SessionHandler(parent=self)
        o_session.save_from_ui()
        o_session.save_to_file()

    def menu_view_load_data_clicked(self):
        self.ui.tabWidget.setCurrentIndex(0)

    def menu_view_normalization_clicked(self):
        self.ui.tabWidget.setCurrentIndex(1)

    def menu_view_normalized_clicked(self):
        self.ui.tabWidget.setCurrentIndex(2)

    def menu_view_binning_clicked(self):
        o_event = GeneralEventHandler(parent=self)
        if o_event.is_step_selected_allowed(step_index_requested=3):
            BinningLauncher(parent=self)

    def menu_view_fitting_clicked(self):
        o_session = SessionHandler(parent=self)
        o_session.save_from_ui()

        o_event = GeneralEventHandler(parent=self)
        if o_event.is_step_selected_allowed(step_index_requested=4):
            FittingLauncher(parent=self)

    def menu_view_strain_mapping_clicked(self):
        o_session = SessionHandler(parent=self)
        o_session.save_from_ui()

        o_event = GeneralEventHandler(parent=self)
        if o_event.is_step_selected_allowed(step_index_requested=5):
            StrainMappingLauncher(parent=self)

    def view_instrument_and_material_settings_clicked(self, state):
        """will make the instrument and material widgets visible, only if we are not working with the
        normalization tab"""

        o_gui = GuiHandler(parent=self)
        tab_selected = o_gui.get_active_tab()

        # ignore it if we are working with the normalization tab
        if tab_selected == DataType.normalization:
            return

        self.ui.instrument_and_material_settings.setVisible(state)

    def log_clicked(self):
        LogLauncher(parent=self)

    def about_clicked(self):
        o_about = AboutLauncher(parent=self)
        o_about.show()

    # TAB 1, 2 and 3  ===========================================================================================
    def tab_widget_changed(self, tab_selected):
        general_event_handler = GeneralEventHandler(parent=self)
        is_step_selected_allowed = general_event_handler.is_step_selected_allowed(step_index_requested=tab_selected)

        if is_step_selected_allowed:
            if tab_selected == 1:  # normalization
                material_instrument_group_visible = False
                o_gui = Step2GuiHandler(parent=self)
                o_gui.update_widgets()
                time_spectra_data = self.data_metadata["time_spectra"]["data"]
                if len(time_spectra_data) == 0:
                    o_gui.enable_xaxis_button(tof_flag=False)
                else:
                    o_gui.enable_xaxis_button(tof_flag=True)
                self.step2_file_index_radio_button_clicked()
                o_plot = Step2Plot(parent=self)
                o_plot.display_bragg_edge()

            elif (tab_selected == 0) or (tab_selected == 2):
                # BraggEdgeElementHandler(parent=self)
                o_plot = Step1Plot(parent=self)
                o_plot.display_general_bragg_edge()
                # self.update_hkl_lambda_d0()

                material_instrument_group_visible = self.ui.action_Instrument_Material_Settings.isChecked()

            elif tab_selected == 3:
                material_instrument_group_visible = False

            self.current_tab = tab_selected
            self.ui.instrument_and_material_settings.setVisible(material_instrument_group_visible)

        else:
            self.ui.tabWidget.setCurrentIndex(self.current_tab)

    def infos_button_clicked(self):
        InfosLauncher(parent=self)

    def infos_window_update(self, data_type=DataType.sample):
        """will update the infos view if this one is active"""
        if self.infos_dict[data_type]:
            self.infos_ui_dict[data_type].setEnabled(True)
        if self.infos_id:
            self.infos_id.update()

    def material_display_clicked(self, status):
        o_gui = Step1GuiHandler(parent=self)
        o_gui.check_time_spectra_widgets()
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    # def material_display_2_clicked(self, status):
    #     self.ui.material_display_checkbox.setChecked(status)
    #     o_gui = Step1GuiHandler(parent=self)
    #     o_gui.check_time_spectra_widgets()
    #     o_plot = Step1Plot(parent=self)
    #     o_plot.display_general_bragg_edge()

    def roi_image_view_changed(self, mouse_selection=True):
        o_plot = Step1Plot(parent=self, data_type=DataType.sample)
        o_plot.display_bragg_edge(mouse_selection=mouse_selection)

    def roi_ob_image_view_changed(self, mouse_selection=True):
        o_plot = Step1Plot(parent=self, data_type=DataType.ob)
        o_plot.display_bragg_edge(mouse_selection=mouse_selection)

    def retrieve_general_infos(self, data_type=DataType.sample):
        if data_type in (DataType.sample, DataType.normalized):
            o_general_infos = RetrieveGeneralFileInfos(parent=self, data_type=data_type)
            o_general_infos.update()

    def retrieve_general_data_infos(self, data_type=DataType.sample):
        if data_type == DataType.sample:
            self.sample_retrieve_general_data_infos()
        elif data_type == DataType.ob:
            self.open_beam_retrieve_general_data_infos()
        elif data_type == DataType.normalized:
            self.normalized_retrieve_general_data_infos()

    def load_data_tab_changed(self, tab_index):
        o_gui = Step1GuiHandler(parent=self)
        o_gui.load_data_tab_changed(tab_index=tab_index)
        if tab_index == 0:
            self.current_data_type = DataType.sample
            self.ui.image_preview.setCurrentIndex(0)
        else:
            self.current_data_type = DataType.ob
            self.ui.image_preview.setCurrentIndex(1)

    def roi_editor_button_clicked(self):
        o_roi_editor = RoiEditor(parent=self)
        o_roi_editor.run()

    def refresh_roi(self, data_type=DataType.sample):
        o_step1_plot = Step1Plot(parent=self, data_type=data_type)
        o_step1_plot.display_bragg_edge()

    def bragg_edge_selection_changed(self):
        o_gui = GuiHandler(parent=self)
        data_type = o_gui.get_active_tab()

        _ui_list = None
        if data_type == DataType.sample:
            _ui_list = self.ui.list_sample
        elif data_type == DataType.ob:
            _ui_list = self.ui.list_open_beam
        else:
            _ui_list = self.ui.list_normalized

        _ui_list.blockSignals(True)
        o_bragg_selection = BraggEdgeSelectionHandler(parent=self, data_type=data_type)
        o_bragg_selection.update_dropdown()

        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=data_type)
        o_retrieve_data_infos.update()

        _ui_list.blockSignals(False)

    # Instrument

    # global load data instruments widgets handler
    def instruments_widgets(self, update_delta_lambda=True):
        if update_delta_lambda:
            self.update_delta_lambda()
        o_data = DataHandler(parent=self)
        o_data.load_time_spectra()
        o_plot = Step1Plot(parent=self)
        o_plot.display_bragg_edge(mouse_selection=False)

    def distance_source_detector_changed(self):
        o_gui = Step1GuiHandler(parent=self)
        o_gui.block_instrument_widgets(status=True)
        self.instruments_widgets()
        o_gui.block_instrument_widgets(status=False)

    def beam_rate_changed(self):
        self.instruments_widgets()

    def detector_offset_changed(self):
        self.instruments_widgets()

    # material - new UI

    def display_material_bragg_peaks_checkbox_clicked(self):
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    def material_tab_changed(self):
        o_material = Material(parent=self)
        o_material.tab_changed()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    ## pre-defined
    def pre_defined_element_dropBox_changed(self, index):
        o_material = MaterialPreDefined(parent=self)
        o_material.update_pre_defined_widgets()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    ## user-defined

    def user_defined_element_name_changed(self, text):
        o_material = Material(parent=self)
        o_material.tab_changed()

    ### method 1

    def fill_fields_with_selected_element_clicked(self, index):
        o_material = MaterialUserDefinedMethod1(parent=self)
        o_material.fill_fields_with_selected_element_clicked()
        o_material = Material(parent=self)
        o_material.check_status_of_all_widgets()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    def user_defined_method1_lattice_changed(self, text):
        o_material = MaterialUserDefinedMethod1(parent=self)
        o_material.lattice_crystal_changed()
        o_material = Material(parent=self)
        o_material.check_status_of_all_widgets()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    def user_defined_method1_crystal_changed(self, text):
        o_material = MaterialUserDefinedMethod1(parent=self)
        o_material.lattice_crystal_changed()
        o_material = Material(parent=self)
        o_material.check_status_of_all_widgets()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    def user_defined_method1_table_changed(self, col, row):
        o_material = Material(parent=self)
        o_material.check_status_of_all_widgets()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    ### method 2
    def user_defined_method2_clear_table_clicked(self):
        o_material = MaterialUserDefinedMethod2(parent=self)
        o_material.clear_user_defined_method2_table()
        o_material = Material(parent=self)
        o_material.check_status_of_all_widgets()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    def user_defined_method2_table_changed(self, col, row):
        o_material = Material(parent=self)
        o_material.check_status_of_all_widgets()
        BraggEdgeElementHandler(parent=self)
        o_plot = Step1Plot(parent=self)
        o_plot.display_general_bragg_edge()

    # TAB 1: Sample and OB Tab =========================================================================================

    def sample_import_button_clicked(self):
        o_event = Step1EventHandler(parent=self, data_type=DataType.sample)
        o_event.import_button_clicked()
        self.load_data_tab_changed(tab_index=0)

    def select_load_data_row(self, data_type=DataType.sample, row=0):
        o_gui = Step1GuiHandler(parent=self)
        o_gui.select_load_data_row(data_type=data_type, row=row)

    def sample_retrieve_general_data_infos(self):
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.sample)
        o_retrieve_data_infos.update()

    def open_beam_retrieve_general_data_infos(self):
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.ob)
        o_retrieve_data_infos.update()

    def sample_list_right_click(self, position):
        o_list_handler = ListDataHandler(parent=self)
        o_list_handler.right_click(position=position)

    def open_beam_import_button_clicked(self):
        o_event = Step1EventHandler(parent=self, data_type=DataType.ob)
        o_event.import_button_clicked()

    def open_beam_list_selection_changed(self):
        if not self.loading_flag:
            o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.ob)
            o_retrieve_data_infos.update()
            self.roi_ob_image_view_changed(mouse_selection=False)
        else:
            self.loading_flag = False

    def time_spectra_import_button_clicked(self):
        o_load = DataHandler(parent=self)
        is_file_selected = o_load.retrieve_time_spectra()
        if is_file_selected:
            o_gui = Step1GuiHandler(parent=self)
            o_gui.check_time_spectra_widgets()
            o_plot = Step1Plot(parent=self)
            o_plot.display_general_bragg_edge()

    def time_spectra_preview_button_clicked(self):
        if self.time_spectra_presenter is None:
            self.time_spectra_presenter = TimeSpectraPresenter(self)

        data_type = self.current_data_type
        file_path = self.data_metadata[data_type]["time_spectra"]["filename"]
        distance_source_detector_m = float(self.ui.distance_source_detector.text())
        detector_offset = float(self.ui.detector_offset.text())

        self.time_spectra_presenter.load_data(file_path, distance_source_detector_m, detector_offset)
        self.time_spectra_presenter.show_view()

    def update_delta_lambda(self):
        o_gui = Step1GuiHandler(parent=self)
        o_gui.update_delta_lambda()

    def roi_algorithm_is_add_clicked(self):
        self.ui.ob_roi_add_button.setChecked(True)
        self.ui.normalized_roi_add_button.setChecked(True)
        self.roi_image_view_changed()

        # update the top plot
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.sample)
        o_retrieve_data_infos.update(add_mean_radio_button_changed=True)

    def roi_algorithm_is_mean_clicked(self):
        self.ui.ob_roi_mean_button.setChecked(True)
        self.ui.normalized_roi_mean_button.setChecked(True)
        self.roi_image_view_changed()

        # update the top plot
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.sample)
        o_retrieve_data_infos.update(add_mean_radio_button_changed=True)

    def ob_roi_algorithm_is_add_clicked(self):
        self.ui.roi_add_button.setChecked(True)
        self.ui.normalized_roi_add_button.setChecked(True)
        self.roi_ob_image_view_changed()

        # update the top plot
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.ob)
        o_retrieve_data_infos.update(add_mean_radio_button_changed=True)

    def ob_roi_algorithm_is_mean_clicked(self):
        self.ui.roi_mean_button.setChecked(True)
        self.ui.normalized_roi_mean_button.setChecked(True)
        self.roi_ob_image_view_changed()

        # update the top plot
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.ob)
        o_retrieve_data_infos.update(add_mean_radio_button_changed=True)

    def file_index_xaxis_button_clicked(self):
        self.data_metadata[DataType.sample][FittingKeys.x_axis] = XAxisMode.file_index_mode
        o_event = Step1EventHandler(parent=self)
        o_event.sample_list_selection_changed()

    def tof_xaxis_button_clicked(self):
        self.data_metadata[DataType.sample][FittingKeys.x_axis] = XAxisMode.tof_mode
        o_event = Step1EventHandler(parent=self)
        o_event.sample_list_selection_changed()

    def lambda_xaxis_button_clicked(self):
        self.data_metadata[DataType.sample][FittingKeys.x_axis] = XAxisMode.lambda_mode
        o_event = Step1EventHandler(parent=self)
        o_event.sample_list_selection_changed()

    def sample_list_selection_changed(self):
        o_event = Step1EventHandler(parent=self)
        o_event.sample_list_selection_changed()

    def ob_file_index_xaxis_button_clicked(self):
        self.data_metadata[DataType.ob][FittingKeys.x_axis] = XAxisMode.file_index_mode
        self.open_beam_list_selection_changed()

    def ob_tof_xaxis_button_clicked(self):
        self.data_metadata[DataType.ob][FittingKeys.x_axis] = XAxisMode.tof_mode
        self.open_beam_list_selection_changed()

    def ob_lambda_xaxis_button_clicked(self):
        self.data_metadata[DataType.ob][FittingKeys.x_axis] = XAxisMode.lambda_mode
        self.open_beam_list_selection_changed()

    def sample_hkl_scrollbar_changed(self, value):
        self.hkl_scrollbar_dict[ScrollBarParameters.value] = value
        o_plot = Step1Plot(parent=self, data_type=DataType.sample)
        o_plot.display_bragg_edge()

    def ob_hkl_scrollbar_changed(self, value):
        self.hkl_scrollbar_dict[ScrollBarParameters.value] = value
        o_plot = Step1Plot(parent=self, data_type=DataType.ob)
        o_plot.display_bragg_edge()

    # TAB 2: Normalization Tab ==================================================================================

    def normalization_manual_roi_changed(self):
        self.ui.normalization_tableWidget.blockSignals(True)
        o_roi = Step2RoiHandler(parent=self)
        o_roi.save_roi()
        o_plot = Step2Plot(parent=self)
        o_plot.update_roi_table()
        o_plot.update_label_roi()
        o_plot.check_error_in_roi_table()
        o_plot.display_bragg_edge()
        self.ui.normalization_tableWidget.blockSignals(False)

    def normalization_row_status_changed(self):
        o_roi = Step2RoiHandler(parent=self)
        o_roi.save_table()
        o_roi.enable_selected_roi()
        o_plot = Step2Plot(parent=self)
        o_plot.display_bragg_edge()
        o_plot.update_label_roi()
        o_gui = Step2GuiHandler(parent=self)
        o_gui.check_run_normalization_button()

    def normalization_row_status_region_type_changed(self, value):
        self.normalization_row_status_changed()
        o_gui = Step2GuiHandler(parent=self)
        o_gui.check_run_normalization_button()

    def normalization_remove_roi_button_clicked(self):
        self.ui.normalization_tableWidget.blockSignals(True)
        o_roi = Step2RoiHandler(parent=self)
        o_roi.remove_roi()
        o_plot = Step2Plot(parent=self)
        o_plot.display_roi()
        o_plot.display_bragg_edge()
        o_gui = Step2GuiHandler(parent=self)
        o_gui.check_run_normalization_button()
        self.ui.normalization_tableWidget.blockSignals(False)

    def normalization_add_roi_button_clicked(self):
        self.ui.normalization_tableWidget.blockSignals(True)
        o_roi = Step2RoiHandler(parent=self)
        o_roi.add_roi()
        o_gui = Step2GuiHandler(parent=self)
        o_gui.check_run_normalization_button()
        self.ui.normalization_tableWidget.blockSignals(False)

    def normalization_button_clicked(self):
        o_norm = Normalization(parent=self)
        norm_worked = o_norm.run_and_export()
        if norm_worked:
            self.ui.tabWidget.setCurrentIndex(2)

    def step2_file_index_radio_button_clicked(self):
        self.data_metadata[DataType.normalization][FittingKeys.x_axis] = XAxisMode.file_index_mode
        o_plot = Step2Plot(parent=self)
        o_plot.display_bragg_edge()

    def step2_tof_radio_button_clicked(self):
        self.data_metadata[DataType.normalization][FittingKeys.x_axis] = XAxisMode.tof_mode
        o_plot = Step2Plot(parent=self)
        o_plot.display_bragg_edge()

    def step2_lambda_radio_button_clicked(self):
        self.data_metadata[DataType.normalization][FittingKeys.x_axis] = XAxisMode.lambda_mode
        o_plot = Step2Plot(parent=self)
        o_plot.display_bragg_edge()

    def normalization_tableWidget_cell_changed(self, row, col):
        o_roi = Step2RoiHandler(parent=self)
        o_roi.save_table()
        o_plot = Step2Plot(parent=self)
        o_plot.display_roi()
        o_plot.check_error_in_roi_table()

    def step2_bragg_edge_selection_changed(self):
        lr = self.bragg_edge_selection
        selection = list(lr.getRegion())

        x_axis = self.current_bragg_edge_x_axis[DataType.normalization]
        left_index = find_nearest_index(array=x_axis, value=selection[0])
        right_index = find_nearest_index(array=x_axis, value=selection[1])
        self.range_files_to_normalized_step2["file_index"] = [left_index, right_index]

    def normalization_moving_average_settings_clicked(self):
        settings = ReductionSettingsHandler(parent=self)
        settings.show()

    # TAB 3: Normalized Tab ==================================================================================

    def normalized_time_spectra_import_button_clicked(self):
        o_load = DataHandler(parent=self, data_type=DataType.normalized)
        is_file_selected = o_load.retrieve_time_spectra()
        if is_file_selected:
            o_gui = Step3GuiHandler(parent=self)
            o_gui.check_time_spectra_widgets()
            o_plot = Step1Plot(parent=self)
            o_plot.display_general_bragg_edge()

    def normalized_time_spectra_preview_button_clicked(self):
        if self.normalized_time_spectra_presenter is None:
            self.normalized_time_spectra_presenter = TimeSpectraPresenter(self)

        data_type = DataType.normalized
        file_path = self.data_metadata[data_type]["time_spectra"]["filename"]
        distance_source_detector_m = float(self.ui.distance_source_detector.text())
        detector_offset = float(self.ui.detector_offset.text())

        self.normalized_time_spectra_presenter.load_data(file_path, distance_source_detector_m, detector_offset)
        self.normalized_time_spectra_presenter.show_view()

    def normalized_import_button_clicked(self):
        o_event = Step3EventHandler(parent=self, data_type=DataType.normalized)
        o_event.import_button_clicked()

    def normalized_retrieve_general_data_infos(self):
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.normalized)
        o_retrieve_data_infos.update()

    def select_normalized_row(self, row=0):
        o_gui = Step3GuiHandler(parent=self)
        o_gui.select_normalized_row(row=row)

    def normalized_list_selection_changed(self):
        QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        if not self.loading_flag:
            o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.normalized)
            o_retrieve_data_infos.update()
            # self.roi_normalized_image_view_changed(mouse_selection=False)
        else:
            self.loading_flag = False
        QApplication.restoreOverrideCursor()

    def roi_normalized_image_view_changed(self, mouse_selection=True):
        o_plot = Step1Plot(parent=self, data_type=DataType.normalized)
        o_plot.display_bragg_edge(mouse_selection=mouse_selection)

    def normalized_roi_algorithm_is_add_clicked(self):
        self.ui.roi_add_button.setChecked(True)
        self.ui.ob_roi_add_button.setChecked(True)
        self.roi_normalized_image_view_changed()

        # update the top plot
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.normalized)
        o_retrieve_data_infos.update(add_mean_radio_button_changed=True)

    def normalized_roi_algorithm_is_mean_clicked(self):
        self.ui.roi_mean_button.setChecked(True)
        self.ui.ob_roi_mean_button.setChecked(True)
        self.roi_normalized_image_view_changed()

        # update the top plot
        o_retrieve_data_infos = RetrieveGeneralDataInfos(parent=self, data_type=DataType.normalized)
        o_retrieve_data_infos.update(add_mean_radio_button_changed=True)

    def normalized_file_index_xaxis_button_clicked(self):
        self.data_metadata[DataType.normalized][FittingKeys.x_axis] = XAxisMode.file_index_mode
        o_event = Step3EventHandler(parent=self)
        o_event.sample_list_selection_changed()

    def normalized_tof_xaxis_button_clicked(self):
        self.data_metadata[DataType.normalized][FittingKeys.x_axis] = XAxisMode.tof_mode
        o_event = Step3EventHandler(parent=self)
        o_event.sample_list_selection_changed()

    def normalized_lambda_xaxis_button_clicked(self):
        self.data_metadata[DataType.normalized][FittingKeys.x_axis] = XAxisMode.lambda_mode
        o_event = Step3EventHandler(parent=self)
        o_event.sample_list_selection_changed()

    def normalized_hkl_scrollbar_changed(self, value):
        self.hkl_scrollbar_dict[ScrollBarParameters.value] = value
        o_plot = Step1Plot(parent=self, data_type=DataType.normalized)
        o_plot.display_bragg_edge()

    # TAB tools

    def rotate_images_clicked(self):
        RotateImages(parent=self)

    def tof_combine_clicked(self):
        TofCombineLauncher(parent=self)

    def tof_binning_clicked(self):
        TofBinningLauncher(parent=self)

    # GENERAL UI =======================================================================================
    def save_session(self):
        o_session = SessionHandler(parent=self)
        o_session.save_from_ui()
        o_session.automatic_save()

    def closeEvent(self, event):
        o_session = SessionHandler(parent=self)
        o_session.save_from_ui()
        o_session.automatic_save()
        self.check_log_file_size()
        logging.info(" #### Leaving iBeatles ####")
        self.close()


def main(args):
    app = QApplication(args)
    app.setStyle("Fusion")
    app.aboutToQuit.connect(clean_up)
    app.setApplicationDisplayName("iBeatles")
    app.setOrganizationName("iBeatles")
    app.setOrganizationDomain("N/A")
    app.setApplicationName("iBeatles")

    # root = os.path.dirname(os.path.realpath(__file__))
    # refresh_image = os.path.join(root, "icons/refresh.png")
    # app.setWindowIcon(QIcon(refresh_image))

    application = MainWindow()
    application.show()
    # sys.exit(app.exec_())
    app.exec_()


def clean_up():
    app = QApplication.instance()
    app.closeAllWindows()



from ..utilities.gui_handler import GuiHandler


def get_tab_selected(parent=None):
    """
    return: 'sample', 'ob', 'normalization' or 'normalized'
    """
    o_gui = GuiHandler(parent=parent)
    return o_gui.get_active_tab()



#!/usr/bin/env python
"""
BinningLauncher class
"""

import numpy as np
import pyqtgraph as pg
from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QMainWindow, QVBoxLayout

from ibeatles import BINNING_LINE_COLOR, DEFAULT_BIN, DEFAULT_ROI, DataType, load_ui
from ibeatles.binning.binning_handler import BinningHandler
from ibeatles.fitting.filling_table_handler import FillingTableHandler
from ibeatles.fitting.fitting_handler import FittingHandler
from ibeatles.fitting.fitting_launcher import FittingLauncher
from ibeatles.session import SessionSubKeys
from ibeatles.utilities import colors


class BinningLauncher:
    def __init__(self, parent=None):
        self.parent = parent

        if self.parent.binning_ui is None:
            binning_window = BinningWindow(parent=parent)
            binning_window.show()
            self.parent.binning_ui = binning_window
            o_binning = BinningHandler(parent=self.parent)
            o_binning.display_image()
        else:
            self.parent.binning_ui.setFocus()
            self.parent.binning_ui.activateWindow()


class BinningWindow(QMainWindow):
    default_bins_settings = {
        "x0": 0,
        "y0": 0,
        "width": 20,
        "height": 20,
        "bins_size": 10,
    }

    image_view = None
    line_view = None
    data = []
    widgets_ui = {
        "x_value": None,
        "y_value": None,
        "intensity_value": None,
        "roi": None,
    }

    def __init__(self, parent=None):
        self.parent = parent
        QMainWindow.__init__(self, parent=parent)
        self.ui = load_ui("ui_binningWindow.ui", baseinstance=self)
        self.setWindowTitle("4. Binning")

        self.load_data()
        self.init_pyqtgraph()
        self.init_widgets()
        self.roi_selection_widgets_modified()
        self.parent.there_is_a_roi = True
        self.parent.data_metadata[DataType.bin]["ui_accessed"] = True
        self.roi_changed_finished()

    def load_data(self):
        self.data = np.array(self.parent.data_metadata["normalized"]["data_live_selection"])

    def init_widgets(self):
        if self.parent.session_dict[DataType.bin][SessionSubKeys.roi]:
            [_, x0, y0, width, height, bin_size] = self.parent.session_dict[DataType.bin][SessionSubKeys.roi]

            self.ui.selection_x0.setText(str(x0))
            self.ui.selection_y0.setText(str(y0))
            self.ui.selection_width.setText(str(width))
            self.ui.selection_height.setText(str(height))
            self.ui.bin_size_horizontalSlider.setValue(int(bin_size))

    def fitting_steps_clicked(self):
        FittingLauncher(parent=self.parent)

    def bin_slider_value_changed(self, new_value):
        self.roi_selection_widgets_modified()

    def roi_changed_finished(self):
        self.roi_selection_widgets_modified()

    def roi_changed(self):
        if self.parent.binning_line_view["ui"]:
            _image_view = self.parent.binning_line_view["image_view"]
            _image_view.removeItem(self.parent.binning_line_view["ui"])
            self.parent.binning_line_view["ui"] = None
        else:
            _image_view = self.parent.binning_line_view["image_view"]

        roi = self.parent.binning_line_view["roi"]
        if len(self.data) == 0:
            return
        region = roi.getArraySlice(self.data, _image_view.imageItem)

        x0 = region[0][0].start
        x1 = region[0][0].stop - 1
        y0 = region[0][1].start
        y1 = region[0][1].stop - 1

        width = np.abs(x0 - x1)
        height = np.abs(y0 - y1)

        self.ui.selection_x0.setText("{}".format(x0))
        self.ui.selection_y0.setText("{}".format(y0))
        self.ui.selection_width.setText("{}".format(width))
        self.ui.selection_height.setText("{}".format(height))

    def init_pyqtgraph(self):
        if len(self.data) == 0:
            status = False
        else:
            status = True

        self.ui.groupBox.setEnabled(status)
        self.ui.groupBox_2.setEnabled(status)
        self.ui.bin_size_horizontalSlider.setEnabled(status)

        pg.setConfigOptions(antialias=True)

        image_view = pg.ImageView(view=pg.PlotItem())
        image_view.ui.roiBtn.hide()
        image_view.ui.menuBtn.hide()
        self.parent.binning_line_view["image_view"] = image_view

        if not self.parent.session_dict[DataType.bin][SessionSubKeys.roi]:
            [_, x0, y0, width, height, _] = self.parent.list_roi[DataType.normalized][0]
            self.parent.session_dict[DataType.bin][SessionSubKeys.roi] = [
                DEFAULT_ROI[0],
                x0,
                y0,
                width,
                height,
                DEFAULT_BIN[-1],
            ]

        binning_roi = self.parent.session_dict[DataType.bin][SessionSubKeys.roi]

        x0 = binning_roi[1]
        y0 = binning_roi[2]
        width = binning_roi[3]
        height = binning_roi[4]

        roi = pg.ROI(
            [int(x0), int(y0)],
            [int(width), int(height)],
            pen=colors.pen_color["0"],
            scaleSnap=True,
        )

        roi.addScaleHandle([1, 1], [0, 0])
        roi.sigRegionChanged.connect(self.roi_changed)
        roi.sigRegionChangeFinished.connect(self.roi_changed_finished)
        self.parent.binning_line_view["roi"] = roi
        image_view.addItem(roi)
        line_view = pg.GraphItem()
        image_view.addItem(line_view)
        self.parent.binning_line_view["ui"] = line_view

        # put everything back into the main GUI
        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(image_view)

        self.ui.left_widget.setLayout(vertical_layout)
        self.ui.left_widget.setVisible(status)

    def get_correct_widget_value(self, ui="", variable_name=""):
        s_variable = str(ui.text())
        if s_variable == "":
            s_variable = str(self.default_bins_settings[variable_name])
            ui.setText(s_variable)
        return int(s_variable)

    def roi_selection_widgets_modified(self):
        if len(self.data) == 0:
            return

        QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        self.parent.table_dictionary = {}

        if self.parent.binning_line_view["ui"]:
            _image_view = self.parent.binning_line_view["image_view"]
            _image_view.removeItem(self.parent.binning_line_view["ui"])
            self.parent.binning_line_view["ui"] = None

        x0 = self.get_correct_widget_value(ui=self.ui.selection_x0, variable_name="x0")
        y0 = self.get_correct_widget_value(ui=self.ui.selection_y0, variable_name="y0")
        width = self.get_correct_widget_value(ui=self.ui.selection_width, variable_name="width")
        height = self.get_correct_widget_value(ui=self.ui.selection_height, variable_name="height")
        bin_size = self.ui.bin_size_horizontalSlider.value()
        self.ui.bin_size_label.setText(str(bin_size))

        # self.parent.binning_bin_size = bin_size
        self.parent.binning_done = True

        self.parent.binning_line_view["roi"].setPos([x0, y0], update=False, finish=False)
        self.parent.binning_line_view["roi"].setSize([width, height], update=False, finish=False)

        pos_adj_dict = self.calculate_matrix_of_pixel_bins(bin_size=bin_size, x0=x0, y0=y0, width=width, height=height)

        pos = pos_adj_dict["pos"]
        adj = pos_adj_dict["adj"]

        line_color = BINNING_LINE_COLOR
        lines = np.array(
            [line_color for n in np.arange(len(pos))],
            dtype=[
                ("red", np.ubyte),
                ("green", np.ubyte),
                ("blue", np.ubyte),
                ("alpha", np.ubyte),
                ("width", float),
            ],
        )

        self.parent.binning_line_view["pos"] = pos
        self.parent.binning_line_view["adj"] = adj
        self.parent.binning_line_view["pen"] = lines

        self.update_binning_bins()

        if self.parent.fitting_ui:
            self.parent.fitting_ui.update_selected_bins_plot()
            self.parent.fitting_ui.check_status_widgets()

            o_table = FittingHandler(parent=self.parent.fitting_ui, grand_parent=self.parent)
            o_table.create_table_dictionary()

            # self.parent.fitting_ui.selection_in_value_table_of_rows_cell_clicked(-1, -1)

            o_handler = FittingHandler(parent=self.parent.fitting_ui, grand_parent=self.parent)
            o_handler.display_roi()
            o_handler.display_locked_active_bins()

            o_fitting = FillingTableHandler(parent=self.parent.fitting_ui, grand_parent=self.parent)
            o_fitting.fill_table()

        self.record_roi()
        QApplication.restoreOverrideCursor()

    def update_binning_bins(self):
        """
        this method takes from the parent file the information necessary to display the selection with
        bins in the binning window
        """

        pos = self.parent.binning_line_view["pos"]
        adj = self.parent.binning_line_view["adj"]
        lines = self.parent.binning_line_view["pen"]

        line_view_binning = pg.GraphItem()
        self.parent.binning_line_view["image_view"].addItem(line_view_binning)
        line_view = line_view_binning
        line_view.setData(pos=pos, adj=adj, pen=lines, symbol=None, pxMode=False)

        self.parent.binning_line_view["ui"] = line_view
        self.record_roi()

    def calculate_matrix_of_pixel_bins(self, bin_size=2, x0=0, y0=0, width=20, height=20):
        pos_adj_dict = {}

        nbr_height_bins = float(height) / float(bin_size)
        real_height = y0 + int(nbr_height_bins) * int(bin_size)

        nbr_width_bins = float(width) / float(bin_size)
        read_width = x0 + int(nbr_width_bins) * int(bin_size)

        # pos (each matrix is one side of the lines)
        pos = []
        adj = []

        # vertical lines
        x = x0
        index = 0
        while x <= x0 + width:
            one_edge = [x, y0]
            other_edge = [x, real_height]
            pos.append(one_edge)
            pos.append(other_edge)
            adj.append([index, index + 1])
            x += bin_size
            index += 2

        # horizontal lines
        y = y0
        while y <= y0 + height:
            one_edge = [x0, y]
            other_edge = [read_width, y]
            pos.append(one_edge)
            pos.append(other_edge)
            adj.append([index, index + 1])
            y += bin_size
            index += 2

        pos_adj_dict["pos"] = np.array(pos)
        pos_adj_dict["adj"] = np.array(adj)

        return pos_adj_dict

    def closeEvent(self, event=None):
        self.parent.binning_ui = None

        if len(self.data) > 0:
            self.record_roi()

        else:
            # reset everything if we quit with no data plotted
            binning_line_view = {
                "ui": None,
                "pos": None,
                "adj": None,
                "pen": None,
                "image_view": None,
                "roi": None,
            }
            self.parent.binning_line_view = binning_line_view

    def record_roi(self):
        x0 = int(str(self.ui.selection_x0.text()))
        y0 = int(str(self.ui.selection_y0.text()))
        width = int(str(self.ui.selection_width.text()))
        height = int(str(self.ui.selection_height.text()))
        bin_size = self.ui.bin_size_horizontalSlider.value()
        self.parent.session_dict[DataType.bin][SessionSubKeys.roi] = [
            DEFAULT_ROI[0],
            x0,
            y0,
            width,
            height,
            bin_size,
        ]
        self.parent.session_dict[DataType.bin][SessionSubKeys.bin_size] = bin_size

    def ok_button_clicked(self):
        self.close()



#!/usr/bin/env python
"""
Event handler for the rotate images tool
"""

import os
import shutil

import numpy as np
import pyqtgraph as pg
import scipy
from loguru import logger
from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QFileDialog

from ibeatles import DataType, interact_me_style, normal_style
from ibeatles.session import SessionSubKeys
from ibeatles.utilities.file_handler import FileHandler, retrieve_timestamp_file_name
from ibeatles.utilities.load_files import LoadFiles
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)


class EventHandler:
    def __init__(self, parent=None, top_parent=None):
        self.parent = parent
        self.top_parent = top_parent

    def select_input_folder(self):
        default_path = self.top_parent.session_dict[DataType.sample][SessionSubKeys.current_folder]
        folder = str(
            QFileDialog.getExistingDirectory(
                caption="Select folder containing images to load",
                directory=default_path,
                options=QFileDialog.ShowDirsOnly,
            )
        )
        if folder == "":
            logger.info("User Canceled the selection of folder!")
            return

        list_tif_files = FileHandler.get_list_of_tif(folder=folder)
        if not list_tif_files:
            logger.info("-> folder does not contain any tif file!")
            show_status_message(
                parent=self.parent,
                message=f"Folder {os.path.basename(folder)} does not contain any TIFF files!",
                duration_s=5,
                status=StatusMessageStatus.error,
            )
            return

        self.parent.ui.folder_selected_label.setText(folder)
        logger.info(f"Users selected the folder: {folder}")
        self.parent.list_tif_files = list_tif_files

    def load_data(self):
        if not self.parent.list_tif_files:
            return

        dict = LoadFiles.load_interactive_data(parent=self.parent, list_tif_files=self.parent.list_tif_files)
        self.parent.image_size["height"] = dict["height"]
        self.parent.image_size["width"] = dict["width"]
        self.parent.images_array = dict["image_array"]

        self.parent.integrated_image = np.mean(dict["image_array"], axis=0)

    def display_data(self):
        if not self.parent.list_tif_array:
            return

    def display_rotated_images(self):
        if self.parent.integrated_image is None:
            return

        data = self.parent.integrated_image
        rotation_value = self.parent.ui.rotation_doubleSpinBox.value()

        rotated_data = scipy.ndimage.interpolation.rotate(data, rotation_value)

        _view = self.parent.ui.image_view.getView()
        _view_box = _view.getViewBox()
        _state = _view_box.getState()

        histogram_level = self.parent.histogram_level
        if histogram_level is None:
            self.parent.first_update = True
        else:
            self.parent.first_update = False

        histogram_widget = self.parent.ui.image_view.getHistogramWidget()
        self.parent.histogram_level = histogram_widget.getLevels()

        self.parent.ui.image_view.setImage(rotated_data)

        _view_box.setState(_state)
        if not self.parent.first_update:
            histogram_widget.setLevels(self.parent.histogram_level[0], self.parent.histogram_level[1])

        self.display_grid(data=rotated_data)

    def display_grid(self, data=None):
        [height, width] = np.shape(data)

        pos = []
        adj = []

        # vertical lines
        x = self.parent.grid_size
        index = 0
        while x <= width:
            one_edge = [x, 0]
            other_edge = [x, height]
            pos.append(one_edge)
            pos.append(other_edge)
            adj.append([index, index + 1])
            x += self.parent.grid_size
            index += 2

        # horizontal lines
        y = self.parent.grid_size
        while y <= height:
            one_edge = [0, y]
            other_edge = [width, y]
            pos.append(one_edge)
            pos.append(other_edge)
            adj.append([index, index + 1])
            y += self.parent.grid_size
            index += 2

        pos = np.array(pos)
        adj = np.array(adj)

        line_color = (0, 255, 0, 255, 0.5)
        lines = np.array(
            [line_color for n in np.arange(len(pos))],
            dtype=[
                ("red", np.ubyte),
                ("green", np.ubyte),
                ("blue", np.ubyte),
                ("alpha", np.ubyte),
                ("width", float),
            ],
        )

        # remove old line_view
        if self.parent.ui.line_view:
            self.parent.ui.image_view.removeItem(self.parent.ui.line_view)
        line_view = pg.GraphItem()
        self.parent.ui.image_view.addItem(line_view)
        line_view.setData(pos=pos, adj=adj, pen=lines, symbol=None, pxMode=False)
        self.parent.ui.line_view = line_view

    def check_widgets(self):
        # enable the slider if there is something to rotate
        if self.parent.integrated_image is not None:
            enable_group_widgets = True

            # enable the process button if the slider is not at zero and there are data loaded
            if float(self.parent.ui.rotation_doubleSpinBox.value()) == 0.0:
                enable_button = False
                self.parent.ui.export_button.setStyleSheet(normal_style)
            else:
                enable_button = True
                self.parent.ui.export_button.setStyleSheet(interact_me_style)

            self.parent.ui.export_button.setEnabled(enable_button)
            self.parent.ui.select_folder_pushButton.setStyleSheet(normal_style)
            self.parent.ui.rotation_doubleSpinBox.setStyleSheet(interact_me_style)

        else:
            enable_group_widgets = False

        self.parent.ui.rotation_angle_groupBox.setEnabled(enable_group_widgets)

    def select_output_folder(self):
        folder = os.path.dirname(self.parent.ui.folder_selected_label.text())
        output_folder = str(QFileDialog.getExistingDirectory(caption="Select output folder ...", directory=folder))

        if not output_folder:
            logger.info(" User cancel rotating the images")
            return

        return output_folder

    def rotate_data(self, output_folder=None):
        if output_folder is None:
            return

        logger.info("Rotating normalized images")
        angle = self.parent.ui.rotation_doubleSpinBox.value()
        data = self.parent.images_array
        import_folder_name = self.parent.ui.folder_selected_label.text()
        list_tiff_files = self.parent.list_tif_files

        QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        full_output_folder_name = EventHandler._create_full_output_folder_name(
            angle=angle, import_folder=import_folder_name, output_folder=output_folder
        )

        nbr_images = len(data)
        self.parent.eventProgress.setMinimum(0)
        self.parent.eventProgress.setMaximum(nbr_images)
        self.parent.eventProgress.setValue(0)
        self.parent.eventProgress.setVisible(True)

        for _index, _data in enumerate(self.parent.images_array):
            rotated_data = scipy.ndimage.interpolation.rotate(_data, angle)

            _input_file = os.path.basename(list_tiff_files[_index])
            new_filename = os.path.join(full_output_folder_name, _input_file)

            EventHandler._save_image(filename=new_filename, data=rotated_data)

            self.parent.eventProgress.setValue(_index + 1)
            QApplication.processEvents()

        # export time stamp data
        EventHandler.export_time_spectra_file(output_folder=full_output_folder_name, input_folder=import_folder_name)

        self.parent.eventProgress.setVisible(False)
        return full_output_folder_name

    @staticmethod
    def export_time_spectra_file(output_folder=None, input_folder=None):
        """use the time spectra file from the first folder selected and export it to the output folder"""
        logger.info("Export time spectra file:")

        # retrieve full path of the time spectra file from first folder selected
        time_spectra_filename = retrieve_timestamp_file_name(folder=input_folder)

        logger.info(f" - time spectra file: {time_spectra_filename}")
        logger.info(f" - to output folder: {output_folder}")

        # copy that spectra file to final location
        shutil.copy(time_spectra_filename, output_folder)

    @staticmethod
    def _create_full_output_folder_name(angle=0.0, import_folder=None, output_folder=None):
        """use the angle value and the folder name of the input data to create the
        output folder name

        ex:
            angle = -1.5
            input_folder = "/users/folder/data/my_input_folder/
            output_folder = "/users/folder/my_output_folder

            output_folder = "/users/folder/my_output_folder/my_input_folder_rotated_by_minus_1_5"
        """
        # define and create output folder file name
        if angle > 0:
            str_rotation_value = f"{angle}"
        else:
            str_rotation_value = f"minus_{np.abs(angle)}"
        str_rotation_value = f"rotated_by_{str_rotation_value}"

        str_rotation_value_parsed = str_rotation_value.split(".")
        new_rotation_value = "_".join(str_rotation_value_parsed)

        input_folder_name = os.path.basename(import_folder)
        full_output_folder_name = os.path.join(output_folder, f"{input_folder_name}_{new_rotation_value}")
        FileHandler.make_or_reset_folder(folder_name=full_output_folder_name)
        logger.info(f" Created folder {full_output_folder_name}")
        return full_output_folder_name

    @staticmethod
    def _save_image(filename="", data=None):
        if os.path.exists(filename):
            os.remove(filename)

        file_name, file_extension = os.path.splitext(filename)
        if file_extension.lower() in [".tif", ".tiff"]:
            FileHandler.make_tiff(data=data, filename=filename)
        elif file_extension.lower() == ".fits":
            FileHandler.make_fits(data=data, filename=filename)
        else:
            logger.info(f"file format {file_extension} not supported!")
            raise NotImplementedError(f"file format {file_extension} not supported!")



#!/usr/bin/env python
"""
EventHandler class
"""

import numpy as np
from loguru import logger

from ibeatles.fitting.selected_bin_handler import SelectedBinsHandler
from ibeatles.utilities.array_utilities import find_nearest_index


class EventHandler:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def bragg_edge_region_changed(self):
        """Updating the x-axis in all units (file index, tof and lambda"""

        # current xaxis is
        x_axis = self.grand_parent.normalized_lambda_bragg_edge_x_axis

        _lr = self.parent.fitting_lr
        if _lr is None:
            return
        selection = list(_lr.getRegion())

        left_index = find_nearest_index(array=x_axis, value=selection[0])
        right_index = find_nearest_index(array=x_axis, value=selection[1])

        list_selected = [left_index, right_index]
        self.grand_parent.fitting_bragg_edge_linear_selection = list_selected

        # display lambda left and right
        lambda_array = self.grand_parent.data_metadata["time_spectra"]["normalized_lambda"] * 1e10
        _lambda_min = lambda_array[left_index]
        _lambda_max = lambda_array[right_index]

        self.parent.ui.lambda_min_lineEdit.setText("{:4.2f}".format(_lambda_min))
        self.parent.ui.lambda_max_lineEdit.setText("{:4.2f}".format(_lambda_max))

    def min_or_max_lambda_manually_changed(self):
        try:
            min_lambda = float(str(self.parent.ui.lambda_min_lineEdit.text()))
            max_lambda = float(str(self.parent.ui.lambda_max_lineEdit.text()))

            lambda_array = self.grand_parent.data_metadata["time_spectra"]["normalized_lambda"] * 1e10

            left_index = find_nearest_index(array=lambda_array, value=min_lambda)
            right_index = find_nearest_index(array=lambda_array, value=max_lambda)

            self.grand_parent.fitting_bragg_edge_linear_selection = [
                left_index,
                right_index,
            ]

            o_bin_handler = SelectedBinsHandler(parent=self.parent, grand_parent=self.grand_parent)
            o_bin_handler.update_bragg_edge_plot()
        except ValueError:
            logger.info("lambda range not yet defined!")

    def hkl_list_changed(self, hkl):
        bragg_edges_array = self.grand_parent.selected_element_bragg_edges_array
        if bragg_edges_array:
            if str(hkl) == "":
                value = "N/A"
            else:
                hkl_array = self.grand_parent.selected_element_hkl_array
                str_hkl_list = ["{},{},{}".format(_hkl[0], _hkl[1], _hkl[2]) for _hkl in hkl_array]
                hkl_bragg_edges = dict(zip(str_hkl_list, bragg_edges_array))
                value = "{:04.4f}".format(float(hkl_bragg_edges[str(hkl)]))
        else:
            value = "N/A"
        self.parent.ui.bragg_edge_calculated.setText(value)

    def check_widgets(self):
        tab_index = self.parent.ui.tabWidget.currentIndex()
        if tab_index == 0:
            show_plot_radio_buttons = True
        else:
            show_plot_radio_buttons = False

        self.parent.ui.plot_label.setVisible(show_plot_radio_buttons)
        self.parent.ui.active_bins_button.setVisible(show_plot_radio_buttons)
        self.parent.ui.locked_bins_button.setVisible(show_plot_radio_buttons)

    def automatically_select_best_lambda_0_for_that_range(self):
        if not self.parent.ui.automatic_hkl0_checkBox.isChecked():
            return

        if not (str(self.parent.ui.lambda_min_lineEdit.text())):
            return

        lambda_min_selected = float(str(self.parent.ui.lambda_min_lineEdit.text()))
        lambda_max_selected = float(str(self.parent.ui.lambda_max_lineEdit.text()))
        mid_lambda = np.mean([lambda_min_selected, lambda_max_selected])

        bragg_edges_array = self.grand_parent.selected_element_bragg_edges_array

        nearest_index = find_nearest_index(bragg_edges_array, mid_lambda)
        self.parent.ui.hkl_list_ui.setCurrentIndex(nearest_index)

    def automatic_hkl0_selection_clicked(self):
        flag_on = self.parent.ui.automatic_hkl0_checkBox.isChecked()
        self.parent.ui.hkl_list_ui.setEnabled(not flag_on)



#!/usr/bin/env python
"""
Get module
"""

from ibeatles.tools.tof_bin import BinAutoMode, BinMode, StatisticsName
from ibeatles.tools.utilities import CombineAlgorithm, TimeSpectraKeys


class Get:
    def __init__(self, parent=None):
        self.parent = parent

    def auto_bins_currently_activated(self):
        auto_bin_mode = self.bin_auto_mode()
        if auto_bin_mode == BinAutoMode.linear:
            return self.parent.linear_bins
        elif auto_bin_mode == BinAutoMode.log:
            return self.parent.log_bins
        else:
            raise NotImplementedError("Auto bin mode not implemented!")

    def x_axis_selected(self):
        if self.parent.bin_file_index_radioButton.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.bin_tof_radioButton.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.bin_lambda_radioButton.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError("xaxis not implemented in the bin tab!")

    def bin_auto_mode(self):
        if self.parent.ui.auto_log_radioButton.isChecked():
            return BinAutoMode.log
        elif self.parent.ui.auto_linear_radioButton.isChecked():
            return BinAutoMode.linear
        else:
            raise NotImplementedError("auto bin mode not implemented!")

    def bin_mode(self):
        if self.parent.ui.bin_tabWidget.currentIndex() == 0:
            return BinMode.auto
        elif self.parent.ui.bin_tabWidget.currentIndex() == 1:
            return BinMode.manual
        elif self.parent.ui.bin_tabWidget.currentIndex() == 2:
            return BinMode.settings
        else:
            raise NotImplementedError("bin mode not implemented!")

    def current_bins_activated(self):
        """
        Looking at the active auto or manual tab and linear or log to figure out which bins
        are currently used in the displayed
        :return: dictionary of the bins to use. Can be either self.parent.linear_bins, self.parent.log_bins
        or self.parent.manual_bins
        """
        bin_mode = self.bin_mode()
        if bin_mode == BinMode.manual:
            return self.parent.manual_bins
        elif bin_mode == BinMode.auto:
            bin_auto_mode = self.bin_auto_mode()
            if bin_auto_mode == BinAutoMode.linear:
                return self.parent.linear_bins
            elif bin_auto_mode == BinAutoMode.log:
                return self.parent.log_bins
            else:
                raise NotImplementedError("bin auto mode not implemented")
        else:
            raise NotImplementedError("bin mode not implemented!")

    def auto_log_bin_requested(self):
        if self.parent.ui.bin_auto_log_file_index_radioButton.isChecked():
            return self.parent.ui.auto_log_file_index_spinBox.value()
        elif self.parent.ui.bin_auto_log_tof_radioButton.isChecked():
            return self.parent.ui.auto_log_tof_doubleSpinBox.value()
        elif self.parent.ui.bin_auto_log_lambda_radioButton.isChecked():
            return self.parent.ui.auto_log_lambda_doubleSpinBox.value()
        else:
            raise NotImplementedError("auto log bin algorithm not implemented!")

    def bin_add_method(self):
        if self.parent.ui.combine_mean_radioButton.isChecked():
            return CombineAlgorithm.mean
        elif self.parent.ui.combine_median_radioButton.isChecked():
            return CombineAlgorithm.median
        else:
            raise NotImplementedError("Combine algorithm is not implemented!")

    def bin_statistics_plot_requested(self):
        current_index = self.parent.ui.bin_stats_comboBox.currentIndex()
        list_name = [
            StatisticsName.mean,
            StatisticsName.median,
            StatisticsName.std,
            StatisticsName.min,
            StatisticsName.max,
        ]
        return list_name[current_index]

    def current_bin_tab_working_axis(self):
        bin_mode = self.bin_auto_mode()
        if bin_mode == BinAutoMode.log:
            return self.bin_log_axis()
        elif bin_mode == BinAutoMode.linear:
            return self.bin_linear_axis()
        else:
            raise NotImplementedError("type not supported")

    def bin_log_axis(self):
        if self.parent.ui.bin_auto_log_file_index_radioButton.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.ui.bin_auto_log_tof_radioButton.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.ui.bin_auto_log_lambda_radioButton.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError("type not supported")

    def bin_linear_axis(self):
        if self.parent.ui.auto_linear_file_index_radioButton.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.ui.auto_linear_tof_radioButton.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.ui.auto_linear_lambda_radioButton.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError("type not supported")

    def manual_working_row(self, working_item_id=None):
        list_item_id = self.parent.list_of_manual_bins_item
        for _row, item in enumerate(list_item_id):
            if item == working_item_id:
                return _row
        return -1



#!/usr/bin/env python
"""
TOF combine launcher
"""

import logging
import warnings

from qtpy.QtWidgets import QMainWindow

from ibeatles import DataType, load_ui
from ibeatles.tools.tof_combine import SessionKeys
from ibeatles.tools.tof_combine.combine.event_handler import (
    EventHandler as CombineEventHandler,
)
from ibeatles.tools.tof_combine.export.export_images import ExportImages
from ibeatles.tools.tof_combine.initialization import Initialization
from ibeatles.tools.tof_combine.tof_combine_export_launcher import (
    TofCombineExportLauncher,
)
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.tools.utilities.reload.reload import Reload
from ibeatles.tools.utilities.time_spectra import TimeSpectraLauncher
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)

warnings.filterwarnings("ignore")


class TofCombineLauncher:
    def __init__(self, parent=None):
        if parent.binning_ui is None:
            tof_combine_window = TofCombine(parent=parent)
            tof_combine_window.show()
            parent.tof_combining_binning_ui = tof_combine_window

        else:
            parent.binning_ui.setFocus()
            parent.binning_ui.activateWindow()


class TofCombine(QMainWindow):
    output_folder = None

    visualize_flag = False

    # list of folders listed in the combine table
    list_folders = None

    # full path to the top folder selected and used to fill the table
    top_folder = None

    # folder we will use or not
    list_of_folders_status = None

    combine_roi = {"x0": 0, "y0": 0, "width": 200, "height": 200}

    # dictionary that will keep record of the entire UI and used to load and save the session
    session = {
        SessionKeys.list_folders: None,
        SessionKeys.list_folders_status: None,
        SessionKeys.top_folder: None,
        SessionKeys.combine_roi: combine_roi,
    }

    # save info from all the folders
    # {0: {SessionKeys.folder: None,
    #      SessionKeys.data: None,
    #      SessionKeys.list_files: None,
    #      SessionKeys.nbr_files: None,
    #      SessionKeys.use: False,
    #      },
    #  1: ....,
    # }
    dict_data_folders = {}

    combine_image_view = None
    combine_roi_item_id = None

    # time spectra dict
    time_spectra = {
        TimeSpectraKeys.file_name: None,
        TimeSpectraKeys.tof_array: None,
        TimeSpectraKeys.lambda_array: None,
        TimeSpectraKeys.file_index_array: None,
    }

    def __init__(self, parent=None):
        """
        Initialization
        Parameters
        ----------
        """
        QMainWindow.__init__(self, parent=parent)
        self.ui = load_ui("ui_tof_combine.ui", baseinstance=self)
        self.parent = parent
        self.initialization()
        self.setup()
        self.setWindowTitle("TOF combine tool")

    def initialization(self):
        o_init = Initialization(parent=self)
        o_init.all()

    def setup(self):
        distance_source_detector = self.parent.ui.distance_source_detector.text()
        self.ui.distance_source_detector_label.setText(distance_source_detector)

        detector_offset = self.parent.ui.detector_offset.text()
        self.ui.detector_offset_label.setText(detector_offset)

    # combine events
    def visualize_clicked(self):
        o_event = CombineEventHandler(parent=self)
        o_event.visualize_flag_changed()
        self.radio_buttons_of_folder_changed()

    # def check_combine_widgets(self):
    #     o_event = CombineEventHandler(parent=self)
    #     o_event.check_widgets()

    def select_top_folder_button_clicked(self):
        o_event = CombineEventHandler(parent=self, grand_parent=self.parent)
        o_event.select_top_folder()

    def refresh_table_clicked(self):
        o_event = CombineEventHandler(parent=self)
        o_event.refresh_table_clicked()

    def radio_buttons_of_folder_changed(self):
        o_event = CombineEventHandler(parent=self, grand_parent=self.parent)
        if self.visualize_flag:
            self.ui.setEnabled(False)
            o_event.update_list_of_folders_to_use()
            o_event.combine_folders()
            o_event.display_profile()
            self.ui.setEnabled(True)
            self.ui.combine_widget.setEnabled(True)
        o_event.check_widgets()

    def time_spectra_preview_clicked(self):
        TimeSpectraLauncher(parent=self)

    def combine_algorithm_changed(self):
        if self.visualize_flag:
            o_event = CombineEventHandler(parent=self)
            o_event.combine_algorithm_changed()
            o_event.display_profile()

    def combine_xaxis_changed(self):
        o_event = CombineEventHandler(parent=self)
        o_event.display_profile()

    def combine_roi_changed(self):
        if self.visualize_flag:
            o_event = CombineEventHandler(parent=self)
            o_event.combine_roi_changed()
            o_event.display_profile()

    def mouse_moved_in_combine_image_preview(self):
        """Mouse moved in the combine pyqtgraph image preview (top right)"""
        pass

    # export images
    def export_combined_and_binned_images_clicked(self):
        pass
        # o_export = ExportImages(parent=self)
        # o_export.run()

    def combine_clicked(self):
        tof_combine_export_ui = TofCombineExportLauncher(parent=self, grand_parent=self.parent)
        tof_combine_export_ui.show()

    def combine_run(self, data_type_selected=DataType.none):
        self.ui.setEnabled(False)

        show_status_message(
            parent=self,
            message="Combining folders ...",
            status=StatusMessageStatus.working,
        )
        o_event = CombineEventHandler(parent=self)
        o_event.update_list_of_folders_to_use()
        o_event.combine_folders()
        o_export = ExportImages(parent=self, top_parent=self.parent)
        o_export.run()
        output_folder = o_export.output_folder

        show_status_message(
            parent=self,
            message="Combining folders ... Done!",
            status=StatusMessageStatus.ready,
            duration_s=5,
        )
        self.ui.setEnabled(True)
        return output_folder

    def reload_run_in_main_ui(self, data_type_selected=DataType.normalized, output_folder=None):
        o_reload = Reload(parent=self, top_parent=self.parent)
        o_reload.run(data_type=data_type_selected, output_folder=output_folder)

    def closeEvent(self, event):
        logging.info(" #### Leaving combine TOF####")
        self.parent.tof_combine_ui = None
        self.close()



# class TimeSpectraKeys:
#
#     tof_array = 'tof_array'
#     lambda_array = 'lambda_array'
#     file_index_array = 'file_index_array'
#     file_name = 'file_name'
#     counts_array = 'counts_array'

#
# class CombineAlgorithm:
#     """list of algorithm used to combine the folders"""
#
#     mean = 'mean'
#     median = 'median'



import numpy as np


def create_list_of_bins_from_selection(top_row=0, bottom_row=0, left_column=0, right_column=0):
    """
    this will return a list of bins(row,column) from the selection

    example1:
            top_row=0, bottom_row=1, left_column=4, right_column=4
            return ((0,4), (1,4))

    example2:
            top_row=3, bottom_row=5, left_column=4, right_column=5
            return ((3,4), (4,4), (5,4), (3,5), (4,5), (5,5))
    """

    list_bins = []
    for _row in np.arange(top_row, bottom_row + 1):
        for _column in np.arange(left_column, right_column + 1):
            list_bins.append((_row, _column))

    list_bins.sort()

    return list(list_bins)


def create_list_of_surrounding_bins(central_bin=None, full_bin_width=None, full_bin_height=None):
    """
    this will return the list of bins surrounding the central_bin coordinates (row, column)

    bin_width = 6
    bin_height = 5

    example1:
        central_bin = (0,0)
        surrounding_bins = [(1,0), (1,1), (0,1)]

    example2:
        central_bin = (5,0)
        surrounding_bins = [(4,0), (4,1), (5,1), (6,1), (6,0)]

    example3:
        central_bin = (4,5)
        surrounding_bins = [(3,4), (3,5), (3,6), (4,4), (4,6), (5,4), (5,5), (5,6)]
    """
    row = central_bin[0]
    column = central_bin[1]

    left_row_value = np.max([row - 1, 0])
    right_row_value = row if (row + 1) >= full_bin_height else (row + 1)

    top_column_value = np.max([column - 1, 0])
    bottom_column_value = column if (column + 1) >= full_bin_width else (column + 1)

    list_surrounding_bins = []
    for _row in np.arange(left_row_value, right_row_value + 1):
        for _column in np.arange(top_column_value, bottom_column_value + 1):
            new_bin = (_row, _column)
            if new_bin == central_bin:
                continue
            list_surrounding_bins.append(new_bin)

    list_surrounding_bins.sort()
    return list_surrounding_bins


def convert_bins_to_keys(list_of_bins=None, full_bin_height=None):
    """
    this convert into a key (used by the self.grand_parent.kropff_table_dictionary) the list of
    bins values (row, column) using the simple math  key = str(_row + _col * nbr_row)
    """
    list_keys = []
    for _bin in list_of_bins:
        _row, _col = _bin
        str_key = str(_row + _col * full_bin_height)
        list_keys.append(str_key)

    return list_keys



import numpy as np


def is_int(value):
    is_number = True
    try:
        int(value)
    except ValueError:
        is_number = False

    return is_number


def is_nan(value):
    if np.isnan(value):
        return True

    return False


def is_float(value):
    is_number = True
    try:
        float(value)
    except ValueError:
        is_number = False

    return is_number



#!/usr/bin/env python
"""View for Time Spectra"""

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from qtpy.QtWidgets import QMainWindow, QMessageBox, QSizePolicy, QVBoxLayout

from ibeatles.app.utils.ui_loader import load_ui


class TimeSpectraView(QMainWindow):
    def __init__(self, presenter):
        super().__init__()
        self.presenter = presenter
        self.ui = load_ui("time_spectra_view.ui", baseinstance=self)
        self.initialize_view()

    def initialize_view(self):
        graphics_view_layout = QVBoxLayout()
        self.ui.time_spectra_view.setLayout(graphics_view_layout)
        self.ui.time_spectra_plot = MatplotlibView(self)
        graphics_view_layout.addWidget(self.ui.time_spectra_plot)

    def plot_data(self, x_axis, y_axis, x2_axis):
        self.ui.time_spectra_plot.ax1.clear()
        self.ui.time_spectra_plot.ax1.plot(x_axis, y_axis, ".")
        self.ui.time_spectra_plot.ax1.set_xlabel(r"$TOF  (\mu s)$")
        self.ui.time_spectra_plot.ax1.set_ylabel("Counts")
        self.ui.time_spectra_plot.figure.subplots_adjust(top=0.9, left=0.1)
        self.ui.time_spectra_plot.draw()

    def set_window_title(self, title):
        self.setWindowTitle(title)

    def set_text_content(self, content):
        self.ui.time_spectra_text.setText(content)

    def show_error(self, title, message):
        QMessageBox.critical(self, title, message)

    def closeEvent(self, event):
        self.presenter.on_view_closed()
        super().closeEvent(event)


class MatplotlibView(FigureCanvas):
    def __init__(self, parent):
        self.fig = Figure()
        self.ax1 = self.fig.add_subplot(111)
        super().__init__(self.fig)
        self.setParent(parent)
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)



#!/usr/bin/env python
"""
Export configuration
"""

import os

from loguru import logger
from qtpy.QtWidgets import QFileDialog

from ibeatles import DataType, FileType
from ibeatles.fitting import FittingTabSelected
from ibeatles.fitting.get import Get
from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.utilities.file_handler import (
    create_full_export_file_name,
)
from ibeatles.utilities.get import Get as MainGet
from ibeatles.utilities.json_handler import save_json
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)
from ibeatles.utilities.time import get_current_time_in_special_file_name_format


class Export:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def config_for_cli(self, output_folder: str = None):
        o_get = Get(parent=self.parent)
        main_tab_selected = o_get.main_tab_selected()
        if main_tab_selected == FittingTabSelected.march_dollase:
            logger.info("Export in Marche Dollase mode not supported yet!")
            show_status_message(
                parent=self.parent,
                message="Export in Marche Dollase mode not supported yet!",
                status=StatusMessageStatus.error,
                duration_s=10,
            )
            return

        _current_time = get_current_time_in_special_file_name_format()

        output_folder = os.path.abspath(output_folder)
        output_file_name = create_full_export_file_name(
            os.path.join(output_folder, f"config_{_current_time}"), FileType.json
        )

        logger.info(f"Exporting configuration to be used by the command line version (CLI) -> {output_file_name}")

        # o_get = Get(parent=self.parent)
        # strain_mapping_dict = o_get.strain_mapping_dictionary()

        session_dict = self.grand_parent.session_dict

        raw_data_dir = session_dict[DataType.sample][SessionSubKeys.current_folder]
        if session_dict[DataType.sample][SessionSubKeys.list_files]:
            _, raw_data_extension = os.path.splitext(session_dict[DataType.sample][SessionSubKeys.list_files][0])
        else:
            raw_data_extension = None

        open_beam_data_dir = session_dict[DataType.ob][SessionSubKeys.current_folder]
        if session_dict[DataType.ob][SessionSubKeys.list_files]:
            _, open_beam_data_extension = os.path.splitext(session_dict[DataType.ob][SessionSubKeys.list_files][0])
        else:
            open_beam_data_extension = None

        list_normalization_roi = session_dict[DataType.normalization][SessionSubKeys.roi]
        normalization_sample_background = []
        for _roi in list_normalization_roi:
            _state, x0, y0, width, height, _type = _roi
            if (_type == "background") and _state:
                normalization_sample_background.append({"x0": x0, "y0": y0, "width": width, "height": height})

        moving_average_active = session_dict[SessionKeys.reduction][SessionSubKeys.activate]
        moving_average_dimension = session_dict[SessionKeys.reduction][SessionSubKeys.dimension]
        moving_average_size = {
            "y": session_dict[SessionKeys.reduction][SessionSubKeys.size]["y"],
            "x": session_dict[SessionKeys.reduction][SessionSubKeys.size]["x"],
        }
        if moving_average_dimension == "3D":
            moving_average_size["z"] = session_dict[SessionKeys.reduction][SessionSubKeys.size["z"]]

        moving_average_type = session_dict[SessionKeys.reduction][SessionSubKeys.type]

        if session_dict[SessionKeys.reduction][SessionSubKeys.process_order] == "option1":
            processing_order = "Moving average, Normalization"
        else:
            processing_order = "Normalization, Moving average"

        o_get = MainGet(parent=self.grand_parent)
        analysis_material_element = o_get.get_material()

        pixel_binning = {
            "x0": session_dict[SessionKeys.bin][SessionSubKeys.roi][1],
            "y0": session_dict[SessionKeys.bin][SessionSubKeys.roi][2],
            "width": session_dict[SessionKeys.bin][SessionSubKeys.roi][3],
            "height": session_dict[SessionKeys.bin][SessionSubKeys.roi][4],
            "bins_size": session_dict[SessionKeys.bin][SessionSubKeys.roi][5],
        }

        fitting_lambda_range = session_dict[SessionKeys.fitting][SessionSubKeys.lambda_range_index]

        # table_dictionary = self.grand_parent.kropff_table_dictionary
        # logger.info(f"{session_dict[SessionKeys.fitting].keys() =}")
        # print(f"{session_dict[SessionKeys.fitting]['xaxis'] = }")

        session_dict[SessionKeys.fitting][SessionSubKeys.xaxis] = [
            float(x) for x in self.grand_parent.normalized_lambda_bragg_edge_x_axis
        ]
        x_axis = session_dict[SessionKeys.fitting][SessionSubKeys.xaxis]

        lambda_min = x_axis[fitting_lambda_range[0]] * 1e-10
        lambda_max = x_axis[fitting_lambda_range[1]] * 1e-10

        lambda_0 = float(self.parent.ui.bragg_edge_calculated.text())
        strain_mapping_d0 = float(f"{lambda_0 / 2.0:04.4f}")

        quality_threshold = 0.8

        distance_source_detector_in_m = session_dict[SessionKeys.instrument][SessionSubKeys.distance_source_detector]
        detector_offset_in_us = session_dict[SessionKeys.instrument][SessionSubKeys.detector_value]

        normalized_data_dir = os.path.join(output_folder, f"normalized_{_current_time}")
        analysis_results_dir = os.path.join(output_folder, f"analysis_{_current_time}")
        strain_results_dir = os.path.join(output_folder, f"strain_{_current_time}")

        config = {
            "raw_data": {
                "raw_data_dir": raw_data_dir,
                "extension": raw_data_extension,
            },
            "open_beam": {
                "open_beam_data_dir": open_beam_data_dir,
                "extension": open_beam_data_extension,
            },
            "normalization": {
                "sample_background": normalization_sample_background,
                "moving_average": {
                    "active": moving_average_active,
                    "dimension": moving_average_dimension,
                    "size": moving_average_size,
                    "type": moving_average_type,
                },
                "processing_order": processing_order,
            },
            "analysis": {
                "material": {
                    "element": analysis_material_element,
                },
                "pixel_binning": pixel_binning,
                "fitting": {
                    "lambda_min": lambda_min,
                    "lambda_max": lambda_max,
                },
                "strain_mapping": {
                    "d0": strain_mapping_d0,
                    "quality_threshold": quality_threshold,
                    "visualization": {
                        "interpolation_method": "nearest",
                        "colormap": "viridis",
                        "alpha": 0.5,
                        "display_fit_quality": True,
                    },
                    "output_file_config": {
                        "strain_map_format": "png",
                        "fitting_grid_format": "pdf",
                        "figure_dpi": 300,
                        "csv_format": {
                            "delimiter": ",",
                            "include_metadata_header": True,
                            "metadata_comment_char": "#",
                            "na_rep": "null",
                        },
                    },
                    "save_intermediate_results": False,
                },
                "distance_source_detector_in_m": distance_source_detector_in_m,
                "detector_offset_in_us": detector_offset_in_us,
            },
            "output": {
                "normalized_data_dir": normalized_data_dir,
                "analysis_results_dir": analysis_results_dir,
                "strain_results_dir": strain_results_dir,
            },
        }

        save_json(json_file_name=output_file_name, json_dictionary=config)

    def select_output_folder(self):
        working_dir = os.path.dirname(
            os.path.dirname(self.grand_parent.session_dict[DataType.normalized][SessionSubKeys.current_folder])
        )
        output_folder = str(
            QFileDialog.getExistingDirectory(self.parent, "Select where to export the config file ...", working_dir)
        )
        return output_folder



#!/usr/bin/env python
"""
Get (step 2)
"""

import numpy as np

from ibeatles.step2 import KernelType, RegionType


class Get:
    def __init__(self, parent=None):
        self.parent = parent

    def kernel_type(self):
        if self.parent.session_dict["reduction"]["type"] == KernelType.box:
            return KernelType.box
        else:
            return KernelType.gaussian

    def table_number_row(self):
        return self.parent.ui.normalization_tableWidget.rowCount()

    def roi_table_row(self, row=-1):
        if row == -1:
            return []

        # use flag
        _flag_widget = self.parent.ui.normalization_tableWidget.cellWidget(row, 0)
        if _flag_widget is None:
            raise ValueError
        flag = _flag_widget.isChecked()

        # x0
        _item = self.parent.ui.normalization_tableWidget.item(row, 1)
        if _item is None:
            raise ValueError
        x0 = str(_item.text())

        # y0
        _item = self.parent.ui.normalization_tableWidget.item(row, 2)
        if _item is None:
            raise ValueError
        y0 = str(_item.text())

        # width
        _item = self.parent.ui.normalization_tableWidget.item(row, 3)
        if _item is None:
            raise ValueError
        width = str(_item.text())

        # height
        _item = self.parent.ui.normalization_tableWidget.item(row, 4)
        if _item is None:
            raise ValueError
        height = str(_item.text())

        # region type
        _text_widget = self.parent.ui.normalization_tableWidget.cellWidget(row, 5)
        region_type = _text_widget.currentText()

        return [flag, x0, y0, width, height, region_type]

    def status_of_run_normalization_button(self):
        """Conditions to validate the run_normalization button are:
        - have ob loaded
        - if no ob, have at least one ENABLED Background ROI
        """
        data = self.parent.data_metadata["sample"]["data"]
        if not data.any():
            return False

        ob = self.parent.data_metadata["ob"]["data"]
        if isinstance(ob, list):
            if not ob:
                return False

        elif not ob.any():
            number_row = self.table_number_row()
            for _row in np.arange(number_row):
                _roi = self.roi_table_row(row=_row)
                [status, _, _, _, _, region_type] = _roi
                if (status is True) and (region_type == RegionType.background):
                    return True

        if ob.any():
            return True

        return False



#!/usr/bin/env python
"""
Initialization module
"""

import pyqtgraph as pg
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QProgressBar, QVBoxLayout

from ibeatles import (
    ANGSTROMS,
    DELTA,
    LAMBDA,
    MICRO,
    auto_image,
    interact_me_style,
    manual_image,
    more_infos_image,
    settings_image,
    stats_plot_image,
    stats_table_image,
)
from ibeatles.utilities.matplotlibview import MatplotlibView
from ibeatles.utilities.table_handler import TableHandler


class Initialization:
    def __init__(self, parent=None, top_parent=None):
        self.parent = parent
        self.top_parent = top_parent

    def all(self):
        self.pyqtgraph_bin()
        self.plot_widgets()
        self.statusbar()
        self.splitter()
        self.table()
        self.labels()
        self.tab()
        self.combobox()
        self.widgets()

    def setup(self):
        distance_source_detector = self.top_parent.ui.distance_source_detector.text()
        self.parent.ui.distance_source_detector_label.setText(distance_source_detector)

        detector_offset = self.top_parent.ui.detector_offset.text()
        self.parent.ui.detector_offset_label.setText(detector_offset)

    def statusbar(self):
        self.parent.eventProgress = QProgressBar(self.parent.ui.statusbar)
        self.parent.eventProgress.setMinimumSize(20, 14)
        self.parent.eventProgress.setMaximumSize(540, 100)
        self.parent.eventProgress.setVisible(False)
        self.parent.ui.statusbar.addPermanentWidget(self.parent.eventProgress)
        self.parent.setStyleSheet("QStatusBar{padding-left:8px;color:red;font-weight:bold;}")

    def splitter(self):
        self.parent.ui.bin_horizontal_splitter.setSizes([300, 800])
        self.parent.ui.bin_horizontal_splitter.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/horizontal_splitter_handle.png");
                                     }
                                     """)
        self.parent.ui.bin_horizontal_splitter.setHandleWidth(15)

        self.parent.ui.bin_vertical_splitter.setSizes([500, 50])
        self.parent.ui.bin_vertical_splitter.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/vertical_splitter_handle.png");
                                     }
                                     """)
        self.parent.ui.bin_vertical_splitter.setHandleWidth(15)

    def combobox(self):
        list_of_options = ["mean", "median", "std", "min", "max"]
        self.parent.ui.bin_stats_comboBox.blockSignals(True)
        self.parent.ui.bin_stats_comboBox.addItems(list_of_options)
        self.parent.ui.bin_stats_comboBox.blockSignals(False)

    def table(self):
        # bin auto table
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        column_sizes = [40, 35, 60, 115, 115]
        o_table.set_column_sizes(column_sizes=column_sizes)
        column_names = [
            "use?",
            "bin #",
            "file #",
            "tof range (" + MICRO + "s)",
            LAMBDA + " range (" + ANGSTROMS + ")",
        ]
        o_table.set_column_names(column_names=column_names)

        # bin manual table
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        column_sizes = [35, 80, 130, 130]
        o_table.set_column_sizes(column_sizes=column_sizes)
        column_names = column_names[1:]
        o_table.set_column_names(column_names=column_names)

        # statistics
        o_table = TableHandler(table_ui=self.parent.ui.statistics_tableWidget)
        column_names = [
            "bin #",
            "file #",
            "tof range (" + MICRO + "s)",
            LAMBDA + " range (" + ANGSTROMS + ")",
            "mean",
            "median",
            "std",
            "min",
            "max",
        ]
        o_table.set_column_names(column_names=column_names)
        column_sizes = [35, 80, 130, 130, 130, 130, 130, 130, 130]
        o_table.set_column_sizes(column_sizes=column_sizes)

    def labels(self):
        self.parent.ui.bin_auto_log_file_index_radioButton.setText(DELTA + "file_index/file_index")
        self.parent.ui.bin_auto_log_tof_radioButton.setText(DELTA + "tof")
        self.parent.ui.bin_auto_log_lambda_radioButton.setText(DELTA + LAMBDA + "/" + LAMBDA)

        self.parent.ui.auto_linear_file_index_radioButton.setText(DELTA + " file index")
        self.parent.ui.auto_linear_tof_radioButton.setText(DELTA + " tof")
        self.parent.ui.auto_linear_lambda_radioButton.setText(DELTA + LAMBDA)
        self.parent.ui.bin_auto_linear_tof_units_label.setText(MICRO + "s")
        self.parent.ui.bin_auto_linear_lambda_units_label.setText(ANGSTROMS)

    def tab(self):
        self.parent.ui.bin_tabWidget.setTabIcon(0, QIcon(auto_image))
        self.parent.ui.bin_tabWidget.setTabIcon(1, QIcon(manual_image))
        self.parent.ui.stats_tabWidget.setTabIcon(0, QIcon(stats_table_image))
        self.parent.ui.stats_tabWidget.setTabIcon(1, QIcon(stats_plot_image))
        self.parent.ui.bin_bottom_tabWidget.setTabIcon(2, QIcon(settings_image))
        self.parent.ui.image_tabWidget.setCurrentIndex(1)

    def plot_widgets(self):
        graphics_view_layout = QVBoxLayout()
        statistics_plot = MatplotlibView(self.parent)
        graphics_view_layout.addWidget(statistics_plot)
        self.parent.ui.statistics_plot_widget.setLayout(graphics_view_layout)
        self.parent.statistics_plot = statistics_plot

    def pyqtgraph_bin(self):
        # integrated image
        image_view = pg.ImageView(view=pg.PlotItem())
        image_view.ui.roiBtn.hide()
        image_view.ui.menuBtn.hide()
        self.parent.integrated_view = image_view
        # image_view.scene.sigMouseMoved.connect(self.parent.mouse_moved_in_integrated_view)
        layout = QVBoxLayout()
        layout.addWidget(image_view)
        self.parent.ui.integrated_image_widget.setLayout(layout)

        # profile
        bin_view = pg.PlotWidget(title="")
        bin_view.plot()
        self.parent.bin_profile_view = bin_view
        layout = QVBoxLayout()
        layout.addWidget(bin_view)
        self.parent.ui.bin_widget.setLayout(layout)

    def widgets(self):
        self.parent.ui.visualize_auto_bins_axis_generated_pushButton.setIcon(QIcon(more_infos_image))
        self.parent.ui.visualize_auto_bins_axis_generated_pushButton.setToolTip("Display full original bin axis")
        self.parent.ui.select_folder_pushButton.setStyleSheet(interact_me_style)



#!/usr/bin/env python
"""
LoadNormalized class
"""

import os

from ibeatles import DataType
from ibeatles.session import SessionSubKeys
from ibeatles.step1.data_handler import DataHandler
from ibeatles.step1.gui_handler import Step1GuiHandler
from ibeatles.step3.gui_handler import Step3GuiHandler
from ibeatles.utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities


class LoadNormalized:
    def __init__(self, parent=None):
        self.parent = parent
        self.session_dict = parent.session_dict
        self.combine_algo = "sum" if self.parent.ui.roi_add_button.isChecked() else "mean"

    def all(self):
        session_dict = self.session_dict
        data_type = DataType.normalized

        list_normalized_files = self.session_dict[data_type][SessionSubKeys.list_files]
        if list_normalized_files:
            input_folder = self.session_dict[data_type][SessionSubKeys.current_folder]
            o_data_handler = DataHandler(parent=self.parent, data_type=data_type)
            list_normalized_files_fullname = [os.path.join(input_folder, _file) for _file in list_normalized_files]
            o_data_handler.load_files(list_of_files=list_normalized_files_fullname)
            time_spectra_file = session_dict[data_type][SessionSubKeys.time_spectra_filename]
            o_data_handler.load_time_spectra(time_spectra_file=time_spectra_file)
            list_files_selected = session_dict[data_type][SessionSubKeys.list_files_selected]
            self.parent.list_file_selected[data_type] = list_files_selected
            self.parent.list_roi[data_type] = session_dict[data_type][SessionSubKeys.list_rois]
            o_gui = Step1GuiHandler(parent=self.parent, data_type=data_type)
            o_gui.initialize_rois_and_labels()

            self.parent.ui.list_normalized.blockSignals(True)
            for _row_selected in list_files_selected:
                _item = self.parent.ui.list_normalized.item(_row_selected)
                _item.setSelected(True)
            self.parent.ui.list_normalized.blockSignals(False)
            self.parent.normalized_list_selection_changed()

            self.parent.retrieve_general_infos(data_type=data_type)
            self.parent.retrieve_general_data_infos(data_type=data_type)
            self.parent.infos_window_update(data_type=data_type)

            o_gui = Step3GuiHandler(parent=self.parent)
            o_gui.check_widgets()
            o_gui.check_time_spectra_widgets()

            self.parent.image_view_settings[data_type]["state"] = session_dict[data_type][
                SessionSubKeys.image_view_state
            ]
            self.parent.image_view_settings[data_type]["histogram"][self.combine_algo] = session_dict[data_type][
                SessionSubKeys.image_view_histogram
            ]

            o_pyqt = PyqtgraphUtilities(
                parent=self.parent,
                image_view=self.parent.ui.normalized_image_view,
                data_type=data_type,
            )
            o_pyqt.set_state(session_dict[data_type][SessionSubKeys.image_view_state])
            # o_pyqt.reload_histogram_level()

            # combine_algo = 'sum' if self.parent.ui.roi_add_button.isChecked() else 'mean'
            histogram_level = session_dict[data_type][SessionSubKeys.image_view_histogram][self.combine_algo]
            o_pyqt.set_histogram_level(histogram_level=histogram_level)

            # self.parent.list_of_element_index_changed(True, data_type=DataType.normalized)
            self.parent.roi_normalized_image_view_changed()

            self.parent.ui.normalized_splitter.setSizes([20, 450])



#!/usr/bin/env python
"""Fitting routines for Kropff analysis"""

import logging
from typing import Dict, Optional, Tuple

import numpy as np
from lmfit import Model
from lmfit.model import ModelResult

from ibeatles.core.fitting.kropff.models import (
    kropff_high_lambda_transmission,
    kropff_low_lambda_transmission,
    kropff_transmission_model,
)
from ibeatles.core.fitting.utils import remove_invalid_data_points


def fit_bragg_edge_single_pass(
    wavelengths: np.ndarray,
    transmission: np.ndarray,
    initial_parameters: Optional[Dict[str, float]] = None,
    parameter_bounds: Optional[Dict[str, Dict[str, float]]] = None,
) -> Optional[ModelResult]:
    """
    Fit Bragg edge using single pass approach.

    Parameters
    ----------
    wavelengths : np.ndarray
        Wavelength values
    transmission : np.ndarray
        Transmission values
    initial_parameters : Optional[Dict[str, float]], optional
        Initial parameter values for fitting. If None, default values will be used.
        Default is None.
    parameter_bounds : Optional[Dict[str, Dict[str, float]]], optional
        Dictionary containing parameter bounds.
        Format: {param_name: {"min": min_value, "max": max_value}}
        If a parameter or bound is not specified, no constraint will be applied.
        Default is None.

    Returns
    -------
    Optional[ModelResult]
        The lmfit ModelResult object containing all fit results and diagnostics.
        Returns None if fitting fails.
    """
    try:
        # Clean data
        wavelengths, transmission = remove_invalid_data_points(wavelengths, transmission)

        # Create model
        model = Model(kropff_transmission_model)

        # Set up parameters with default values if not provided
        if initial_parameters is None:
            initial_parameters = {
                "a0": 0.1,
                "b0": 0.1,
                "a_hkl": 0.1,
                "b_hkl": 0.1,
                "bragg_edge_wavelength": 4.0,
                "sigma": 0.01,
                "tau": 0.01,
            }

        # Create parameters
        params = model.make_params(**initial_parameters)

        # Apply bounds if provided
        if parameter_bounds:
            for param_name, bounds in parameter_bounds.items():
                if param_name in params:
                    if "min" in bounds:
                        params[param_name].min = bounds["min"]
                    if "max" in bounds:
                        params[param_name].max = bounds["max"]

        # Perform fit
        result = model.fit(transmission, params, wavelength=wavelengths)

        if result.success:
            logging.debug(f"Fit successful: {result.fit_report()}")
        else:
            logging.warning(f"Fit completed but may not be optimal: {result.message}")

        return result

    except Exception as e:
        logging.warning(f"Fitting failed: {str(e)}")
        return None


def fit_bragg_edge_with_refinement(
    wavelengths: np.ndarray,
    transmission: np.ndarray,
    initial_parameters: Optional[Dict[str, float]] = None,
    parameter_bounds: Optional[Dict[str, Dict[str, float]]] = None,
) -> Optional[ModelResult]:
    """
    Fit Bragg edge using multi-step refinement approach.
    This method performs fitting in three steps:
    1. Initial fit with all parameters
    2. Refine a0 and b0 while fixing other parameters
    3. Refine a_hkl and b_hkl while fixing other parameters
    4. Final refinement of edge parameters while fixing decay parameters

    Parameters
    ----------
    wavelengths : np.ndarray
        Wavelength values
    transmission : np.ndarray
        Transmission values
    initial_parameters : Optional[Dict[str, float]], optional
        Initial parameter values for fitting. If None, default values will be used.
        Default is None.
    parameter_bounds : Optional[Dict[str, Dict[str, float]]], optional
        Dictionary containing parameter bounds.
        Format: {param_name: {"min": min_value, "max": max_value}}
        Bounds for sigma and tau must be provided to ensure numerical stability.
        Default is None.

    Returns
    -------
    Optional[ModelResult]
        The lmfit ModelResult object containing all fit results and diagnostics.
        Returns None if fitting fails.
    """
    # Check if required bounds are provided, missing bounds means numerical instability, therefore
    # we raise an error to force user to provide bounds
    if parameter_bounds is None or "sigma" not in parameter_bounds or "tau" not in parameter_bounds:
        raise ValueError("Bounds for 'sigma' and 'tau' must be provided to ensure numerical stability")

    def setup_parameters(model, prev_result, vary_params, bounds=None):
        """Helper function to set up parameters for each step."""
        params = model.make_params()

        # If this is initial setup (no prev_result)
        if prev_result is None:
            for name, value in initial_parameters.items():
                params[name].set(value=value)
        else:
            for name, param in prev_result.params.items():
                params[name].set(value=param.value, vary=name in vary_params)

        # Apply bounds
        if bounds:
            for param_name, param_bounds in bounds.items():
                if param_name in params:
                    for bound_type, value in param_bounds.items():
                        setattr(params[param_name], bound_type, value)

        return params

    try:
        # Clean data
        wavelengths, transmission = remove_invalid_data_points(wavelengths, transmission)

        # Create model
        model = Model(kropff_transmission_model)

        # Set up initial parameters
        if initial_parameters is None:
            initial_parameters = {
                "a0": 0.1,
                "b0": 0.1,
                "a_hkl": 0.1,
                "b_hkl": 0.1,
                "bragg_edge_wavelength": 4.0,
                "sigma": 0.01,
                "tau": 0.01,
            }

        # Initial blind fit with bounds
        initial_params = setup_parameters(
            model,
            None,  # No previous result for initial fit
            vary_params=initial_parameters.keys(),
            bounds=parameter_bounds,
        )

        initial_result = model.fit(
            transmission,
            initial_params,
            wavelength=wavelengths,
        )

        if not initial_result.success:
            logging.warning("Initial fit failed")
            return None

        # Define refinement steps
        refinement_steps = [
            {"vary": ["a0", "b0"]},
            {"vary": ["a_hkl", "b_hkl"]},
            {"vary": ["bragg_edge_wavelength", "sigma", "tau"]},
        ]

        current_result = initial_result
        for step in refinement_steps:
            params = setup_parameters(model, current_result, step["vary"], parameter_bounds)
            current_result = model.fit(transmission, params, wavelength=wavelengths)

        if current_result.success:
            logging.debug(f"Multi-step refinement successful: {current_result.fit_report()}")
        else:
            logging.warning(f"Multi-step refinement completed but may not be optimal: {current_result.message}")

        return current_result

    except Exception as e:
        logging.warning(f"Multi-step refinement failed: {str(e)}")
        return None


def fit_bragg_edge_multi_step(
    wavelengths: np.ndarray,
    transmission: np.ndarray,
    high_lambda_range: Tuple[float, float],
    low_lambda_range: Tuple[float, float],
    initial_parameters: Optional[Dict[str, float]] = None,
    parameter_bounds: Optional[Dict[str, Dict[str, float]]] = None,
) -> Optional[ModelResult]:
    """
    Fit Bragg edge using traditional Kropff multi-step approach.

    Parameters
    ----------
    wavelengths : np.ndarray
        Wavelength values
    transmission : np.ndarray
        Transmission values
    high_lambda_range : Tuple[float, float]
        (min, max) wavelength range for high lambda fitting
    low_lambda_range : Tuple[float, float]
        (min, max) wavelength range for low lambda fitting
    initial_parameters : Optional[Dict[str, float]], optional
        Initial parameter values for fitting. If None, default values will be used.
        Default is None.
    parameter_bounds : Optional[Dict[str, Dict[str, float]]], optional
        Dictionary containing parameter bounds.
        Format: {param_name: {"min": min_value, "max": max_value}}
        Bounds for sigma and tau must be provided to ensure numerical stability.
        Default is None.

    Returns
    -------
    Optional[ModelResult]
        The lmfit ModelResult object containing all fit results and diagnostics.
        Returns None if fitting fails.
    """
    # Validate bounds first
    if parameter_bounds is None or "sigma" not in parameter_bounds or "tau" not in parameter_bounds:
        raise ValueError("Bounds for 'sigma' and 'tau' must be provided to ensure numerical stability")

    try:
        # Clean data
        wavelengths, transmission = remove_invalid_data_points(wavelengths, transmission)

        # Set up initial parameters if not provided
        if initial_parameters is None:
            initial_parameters = {
                "a0": 0.1,
                "b0": 0.1,
                "a_hkl": 0.1,
                "b_hkl": 0.1,
                "bragg_edge_wavelength": 4.0,
                "sigma": 0.01,
                "tau": 0.01,
            }

        # Step 1: Fit high lambda region
        high_lambda_mask = (wavelengths >= high_lambda_range[0]) & (wavelengths <= high_lambda_range[1])
        high_lambda_model = Model(kropff_high_lambda_transmission)

        high_lambda_params = high_lambda_model.make_params(a0=initial_parameters["a0"], b0=initial_parameters["b0"])

        if parameter_bounds and "a0" in parameter_bounds:
            high_lambda_params["a0"].min = parameter_bounds["a0"].get("min")
            high_lambda_params["a0"].max = parameter_bounds["a0"].get("max")
        if parameter_bounds and "b0" in parameter_bounds:
            high_lambda_params["b0"].min = parameter_bounds["b0"].get("min")
            high_lambda_params["b0"].max = parameter_bounds["b0"].get("max")

        high_lambda_result = high_lambda_model.fit(
            transmission[high_lambda_mask],
            high_lambda_params,
            wavelength=wavelengths[high_lambda_mask],
        )

        if not high_lambda_result.success:
            logging.warning("High lambda fit failed")
            return None

        # Step 2: Fit low lambda region using fixed high lambda parameters
        low_lambda_mask = (wavelengths >= low_lambda_range[0]) & (wavelengths <= low_lambda_range[1])
        low_lambda_model = Model(kropff_low_lambda_transmission)

        low_lambda_params = low_lambda_model.make_params(
            a0=high_lambda_result.params["a0"].value,
            b0=high_lambda_result.params["b0"].value,
            a_hkl=initial_parameters["a_hkl"],
            b_hkl=initial_parameters["b_hkl"],
        )

        # Fix high lambda parameters
        low_lambda_params["a0"].vary = False
        low_lambda_params["b0"].vary = False

        if parameter_bounds and "a_hkl" in parameter_bounds:
            low_lambda_params["a_hkl"].min = parameter_bounds["a_hkl"].get("min")
            low_lambda_params["a_hkl"].max = parameter_bounds["a_hkl"].get("max")
        if parameter_bounds and "b_hkl" in parameter_bounds:
            low_lambda_params["b_hkl"].min = parameter_bounds["b_hkl"].get("min")
            low_lambda_params["b_hkl"].max = parameter_bounds["b_hkl"].get("max")

        low_lambda_result = low_lambda_model.fit(
            transmission[low_lambda_mask],
            low_lambda_params,
            wavelength=wavelengths[low_lambda_mask],
        )

        if not low_lambda_result.success:
            logging.warning("Low lambda fit failed")
            return None

        # Step 3: Final full profile fit with fixed decay parameters
        full_model = Model(kropff_transmission_model)

        final_params = full_model.make_params(
            a0=high_lambda_result.params["a0"].value,
            b0=high_lambda_result.params["b0"].value,
            a_hkl=low_lambda_result.params["a_hkl"].value,
            b_hkl=low_lambda_result.params["b_hkl"].value,
            bragg_edge_wavelength=initial_parameters["bragg_edge_wavelength"],
            sigma=initial_parameters["sigma"],
            tau=initial_parameters["tau"],
        )

        # Fix decay parameters
        final_params["a0"].vary = False
        final_params["b0"].vary = False
        final_params["a_hkl"].vary = False
        final_params["b_hkl"].vary = False

        # Apply bounds for edge parameters
        for param_name in ["bragg_edge_wavelength", "sigma", "tau"]:
            if parameter_bounds and param_name in parameter_bounds:
                if "min" in parameter_bounds[param_name]:
                    final_params[param_name].min = parameter_bounds[param_name]["min"]
                if "max" in parameter_bounds[param_name]:
                    final_params[param_name].max = parameter_bounds[param_name]["max"]

        final_result = full_model.fit(transmission, final_params, wavelength=wavelengths)

        if final_result.success:
            logging.debug(f"Multi-step fit successful: {final_result.fit_report()}")
        else:
            logging.warning(f"Multi-step fit completed but may not be optimal: {final_result.message}")

        return final_result

    except Exception as e:
        logging.warning(f"Multi-step fit failed: {str(e)}")
        return None



#!/usr/bin/env python
"""
Bragg edge element handler
"""

from neutronbraggedge.braggedge import BraggEdge

from ibeatles import (
    MATERIAL_BRAGG_PEAK_TO_DISPLAY_AT_THE_SAME_TIME,
    ScrollBarParameters,
)
from ibeatles.session import MaterialMode
from ibeatles.utilities.gui_handler import GuiHandler


class BraggEdgeElementHandler:
    bragg_edges_array = []

    def __init__(self, parent=None):
        self.parent = parent

        table_ui_dict = {
            MaterialMode.pre_defined: self.parent.ui.pre_defined_tableWidget,
            MaterialMode.custom_method1: self.parent.ui.method1_tableWidget,
            MaterialMode.custom_method2: self.parent.ui.method2_tableWidget,
        }

        o_gui = GuiHandler(parent=self.parent)

        material_active_tab = o_gui.get_material_active_tab()
        table_ui = table_ui_dict[material_active_tab]

        table_data, column_names = o_gui.collect_table_data(table_ui=table_ui)
        list_hkl, list_lambda = BraggEdgeElementHandler.extract_hkl_lambda_from_table(table_data=table_data)

        self.parent.selected_element_bragg_edges_array = list_lambda
        self.parent.selected_element_hkl_array = list_hkl
        self.parent.selected_element_name = "FIXME"

        self.reset_scroll_bar_in_bottom_right_plot()

    @staticmethod
    def extract_hkl_lambda_from_table(table_data=None):
        """
        using the table from the pre-defined, method1 and method2, will extract the rows into hkl and lambda list
        """
        list_hkl = []
        list_lambda = []
        for _index in table_data.keys():
            _row_entry = table_data[_index]
            _row_hkl = [_row_entry["h"], _row_entry["k"], _row_entry["l"]]
            list_hkl.append(_row_hkl)

            if "d0" in _row_entry.keys():
                if _row_entry["d0"] is None:
                    continue
                list_lambda.append(float(_row_entry["d0"]) * 2)
            else:
                list_lambda.append(_row_entry["\u03bb\u2090"])

        return list_hkl, list_lambda

    def reset_scroll_bar_in_bottom_right_plot(self):
        _selected_element_bragg_edges_array = self.parent.selected_element_bragg_edges_array
        nbr_hkl_in_list = len(_selected_element_bragg_edges_array)
        scrollbar_max = nbr_hkl_in_list - MATERIAL_BRAGG_PEAK_TO_DISPLAY_AT_THE_SAME_TIME

        self.parent.hkl_scrollbar_dict = {
            ScrollBarParameters.maximum: scrollbar_max,
            ScrollBarParameters.value: scrollbar_max,
        }


class BraggEdgeElementCalculator:
    element_name = None
    lattice_value = None
    crystal_structure = None

    hkl_array = None
    lambda_array = None
    d0_array = None

    def __init__(self, element_name=None, lattice_value=None, crystal_structure=None):
        self.element_name = element_name
        self.lattice_value = lattice_value
        self.crystal_structure = crystal_structure

    def run(self):
        _element_dictionary = {
            "name": self.element_name,
            "lattice": self.lattice_value,
            "crystal_structure": self.crystal_structure,
        }

        _handler = BraggEdge(new_material=[_element_dictionary])

        self.hkl_array = _handler.hkl[self.element_name]
        self.lambda_array = _handler.bragg_edges[self.element_name]
        self.d0_array = [_value / 2.0 for _value in self.lambda_array]



#!/usr/bin/env python
"""
Table Fitting Story Dictionary Handler
"""

import collections

import numpy as np

from ibeatles.utilities.status import Status


class TableFittingStoryDictionaryHandler:
    story_1 = [
        ["a1", "a6"],
        ["a2", "a5"],
        ["d_spacing", "sigma", "alpha"],
        ["a1", "a2", "a5", "a6"],
        ["d_spacing", "sigma", "alpha"],
        ["d_spacing", "sigma", "alpha", "a1", "a2", "a5", "a6"],
    ]

    list_widget_tag = ["d_spacing", "sigma", "alpha", "a1", "a2", "a5", "a6"]

    init_entry = {
        "d_spacing": False,
        "sigma": False,
        "alpha": False,
        "a1": False,
        "a2": False,
        "a5": False,
        "a6": False,
        "progress": Status.not_run_yet,
    }

    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def initialize_table(self):
        table_fitting_story_dictionary = collections.OrderedDict()

        for row in np.arange(6):
            table_fitting_story_dictionary[row] = {
                "d_spacing": False,
                "sigma": False,
                "alpha": False,
                "a1": False,
                "a2": False,
                "a5": False,
                "a6": False,
                "progress": Status.not_run_yet,
            }

        for _index, _story in enumerate(self.story_1):
            for _variable in _story:
                table_fitting_story_dictionary[_index][_variable] = True

        self.grand_parent.table_fitting_story_dictionary = table_fitting_story_dictionary

    def move_entry(self, current_index_row=0, direction="up"):
        table_fitting_story_dictionary = self.grand_parent.table_fitting_story_dictionary

        if direction == "up":
            new_index_row = current_index_row - 1
        else:
            new_index_row = current_index_row + 1

        tmp_entry = table_fitting_story_dictionary[new_index_row]
        table_fitting_story_dictionary[new_index_row] = table_fitting_story_dictionary[current_index_row]
        table_fitting_story_dictionary[current_index_row] = tmp_entry

        self.grand_parent.table_fitting_story_dictionary = table_fitting_story_dictionary

    def remove_entry(self, index_to_remove=0):
        table_fitting_story_dictionary = self.grang_parent.table_fitting_story_dictionary
        # nbr_entry = len(table_fitting_story_dictionary)

        new_table_fitting_story_dictionary = collections.OrderedDict()
        new_index = 0
        for _index in table_fitting_story_dictionary.keys():
            if _index == index_to_remove:
                continue

            new_table_fitting_story_dictionary[new_index] = table_fitting_story_dictionary[_index]
            new_index += 1

        self.grand_parent.table_fitting_story_dictionary = new_table_fitting_story_dictionary

    def add_entry(self, index_to_add=1):
        table_fitting_story_dictionary = self.grand_parent.table_fitting_story_dictionary
        nbr_entry = len(table_fitting_story_dictionary)

        new_table_fitting_story_dictionary = collections.OrderedDict()
        if table_fitting_story_dictionary == {}:
            new_table_fitting_story_dictionary[0] = self.init_entry

        else:
            new_index = 0
            if index_to_add == nbr_entry:  # if new row is at the last row
                new_table_fitting_story_dictionary = table_fitting_story_dictionary
                new_table_fitting_story_dictionary[index_to_add] = self.init_entry

            else:  # for all other rows
                for _index in table_fitting_story_dictionary.keys():
                    if _index == index_to_add:
                        new_table_fitting_story_dictionary[new_index] = self.init_entry
                        new_index += 1

                    new_table_fitting_story_dictionary[new_index] = table_fitting_story_dictionary[_index]
                    new_index += 1

        self.grand_parent.table_fitting_story_dictionary = new_table_fitting_story_dictionary



#!/usr/bin/env python
"""
Get class for fitting
"""

import numpy as np

from ibeatles.fitting.kropff.fitting_functions import (
    kropff_bragg_peak_tof,
    kropff_high_lambda,
    kropff_low_lambda,
)
from ibeatles.fitting.kropff.get import Get as KropffGet
from ibeatles.utilities.array_utilities import find_nearest_index

from ..utilities.table_handler import TableHandler
from . import FittingTabSelected, KropffTabSelected


class Get:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def main_tab_selected(self):
        tab_selected_index = self.parent.ui.tabWidget.currentIndex()
        if tab_selected_index == 0:
            return FittingTabSelected.march_dollase
        elif tab_selected_index == 1:
            return FittingTabSelected.kropff
        else:
            raise IndexError("Fitting algorithm not implemented")

    def fitting_bragg_edge_linear_selection(self):
        if len(self.grand_parent.fitting_bragg_edge_linear_selection) == 0:
            linear_region_left_index = 0
            x_axis = self.grand_parent.normalized_lambda_bragg_edge_x_axis
            linear_region_right_index = len(x_axis) - 1
            self.grand_parent.fitting_bragg_edge_linear_selection = [
                linear_region_left_index,
                linear_region_right_index,
            ]
            return [linear_region_left_index, linear_region_right_index]

        else:
            return self.grand_parent.fitting_bragg_edge_linear_selection

    def y_axis_fitted_for_given_rows_selected(self):
        o_get = KropffGet(parent=self.parent)
        table_ui_selected = o_get.kropff_table_ui_selected()
        row_selected = self.row_selected_for_this_table_ui(table_ui=table_ui_selected)
        table_dictionary = self.grand_parent.kropff_table_dictionary
        kropff_tab_selected = o_get.kropff_tab_selected()

        list_of_yaxis_fitted = []
        xaxis = []

        for _row in row_selected:
            _row = str(_row)
            _bin_entry = table_dictionary[_row]

            if _bin_entry["fitted"][kropff_tab_selected]["yaxis"] is None:
                xaxis, _yaxis_fitted = self.calculate_yaxis_fitted_using_fitted_parameters(
                    kropff_tab_selected=kropff_tab_selected, row=_row
                )

                if _yaxis_fitted is None:
                    return [], []

                else:
                    table_dictionary[_row]["fitted"][kropff_tab_selected]["xaxis"] = xaxis
                    table_dictionary[_row]["fitted"][kropff_tab_selected]["yaxis"] = _yaxis_fitted
                    list_of_yaxis_fitted.append(_yaxis_fitted)

            else:
                xaxis = _bin_entry["fitted"][kropff_tab_selected]["xaxis"]
                list_of_yaxis_fitted.append(_bin_entry["fitted"][kropff_tab_selected]["yaxis"])

        return xaxis, list_of_yaxis_fitted

    def calculate_yaxis_fitted_using_fitted_parameters(self, kropff_tab_selected=KropffTabSelected.high_tof, row="0"):
        table_dictionary = self.grand_parent.kropff_table_dictionary
        full_xaxis = table_dictionary[row]["xaxis"]

        a0 = table_dictionary[row]["a0"]["val"]
        b0 = table_dictionary[row]["b0"]["val"]
        if np.isnan(a0):
            return [], []

        if kropff_tab_selected == KropffTabSelected.high_tof:
            right = table_dictionary[row]["bragg peak threshold"]["right"]
            nearest_index = find_nearest_index(full_xaxis, right)
            xaxis = full_xaxis[nearest_index:-1]
            yaxis_fitted = kropff_high_lambda(xaxis, a0, b0)

            return xaxis, yaxis_fitted

        else:
            ahkl = table_dictionary[row]["ahkl"]["val"]
            bhkl = table_dictionary[row]["bhkl"]["val"]
            if np.isnan(ahkl):
                return [], []

            if kropff_tab_selected == KropffTabSelected.low_tof:
                left = table_dictionary[row]["bragg peak threshold"]["left"]
                nearest_index = find_nearest_index(full_xaxis, left)
                xaxis = full_xaxis[: nearest_index + 1]
                yaxis_fitted = kropff_low_lambda(xaxis, a0, b0, ahkl, bhkl)

                return xaxis, yaxis_fitted

            elif kropff_tab_selected == KropffTabSelected.bragg_peak:
                lambda_hkl = table_dictionary[row]["lambda_hkl"]["val"]
                tau = table_dictionary[row]["tau"]["val"]
                sigma = table_dictionary[row]["sigma"]["val"]

                xaxis = full_xaxis
                yaxis_fitted = kropff_bragg_peak_tof(xaxis, a0, b0, ahkl, bhkl, lambda_hkl, sigma, tau)

                return xaxis, yaxis_fitted

            else:
                raise NotImplementedError

    def y_axis_and_x_axis_for_given_rows_selected(self):
        o_get = KropffGet(parent=self.parent)
        table_ui_selected = o_get.kropff_table_ui_selected()
        row_selected = self.row_selected_for_this_table_ui(table_ui=table_ui_selected)
        table_dictionary = self.grand_parent.kropff_table_dictionary

        # data_2d = np.array(self.grand_parent.data_metadata['normalized']['data'])
        data_2d = self.grand_parent.data_metadata["normalized"]["data"]
        if type(data_2d) is not type(np.array):
            data_2d = np.array(data_2d)

        self.grand_parent.data_metadata["normalized"]["data"] = data_2d

        # index of selection in bragg edge plot
        [left_index, right_index] = self.fitting_bragg_edge_linear_selection()

        list_of_yaxis = []

        if len(row_selected) == 0:
            return [], []

        for _row in row_selected:
            _bin_entry = table_dictionary[str(_row)]

            if _bin_entry["yaxis"] is None:
                _bin_x0 = int(_bin_entry["bin_coordinates"]["x0"])
                _bin_x1 = int(_bin_entry["bin_coordinates"]["x1"])
                _bin_y0 = int(_bin_entry["bin_coordinates"]["y0"])
                _bin_y1 = int(_bin_entry["bin_coordinates"]["y1"])

                yaxis = data_2d[
                    left_index:right_index,
                    _bin_x0:_bin_x1,
                    _bin_y0:_bin_y1,
                ]  # noqa: E124

                yaxis = np.nanmean(yaxis, axis=1)
                yaxis = np.array(np.nanmean(yaxis, axis=1), dtype=float)
                _bin_entry["yaxis"] = yaxis
                self.grand_parent.kropff_table_dictionary[str(_row)] = _bin_entry

                # index of selection in bragg edge plot
                # [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection
                full_x_axis = self.parent.bragg_edge_data["x_axis"]
                xaxis = np.array(full_x_axis[left_index:right_index], dtype=float)
                _bin_entry["xaxis"] = xaxis

            else:
                yaxis = _bin_entry["yaxis"]
                xaxis = _bin_entry["xaxis"]

            list_of_yaxis.append(yaxis)

        return list_of_yaxis, xaxis

    def row_selected_for_this_table_ui(self, table_ui=None):
        if table_ui is None:
            return []

        o_table = TableHandler(table_ui=table_ui)
        row_selected = o_table.get_rows_of_table_selected()
        return row_selected

    def is_automatic_bragg_peak_threshold_finder_activated(self):
        return self.parent.ui.kropff_automatic_bragg_peak_threshold_finder_checkBox.isChecked()



#!/usr/bin/env python
"""
Material class
"""

import numpy as np
from neutronbraggedge.braggedge import BraggEdge

from ibeatles.utilities.bragg_edge_element_handler import BraggEdgeElementCalculator
from ibeatles.utilities.check import is_float, is_int
from ibeatles.utilities.table_handler import TableHandler


class Material:
    def __init__(self, parent=None):
        self.parent = parent

    def tab_changed(self):
        self.check_status_of_all_widgets()

    def check_status_of_all_widgets(self):
        """
        check the status of the widgets
        """
        top_tab_index = self.parent.ui.material_top_tabWidget.currentIndex()
        if top_tab_index == 0:
            return

        if top_tab_index == 1:
            element_name = self.parent.ui.user_defined_element_name.text()
            if element_name.strip() == "":
                self.parent.ui.material_custom_tabWidget.setEnabled(False)
                self.parent.ui.user_defined_name_error.setVisible(True)

            else:
                self.parent.ui.material_custom_tabWidget.setEnabled(True)
                self.parent.ui.user_defined_name_error.setVisible(False)

                custom_tab_index = self.parent.ui.material_custom_tabWidget.currentIndex()
                if custom_tab_index == 0:  # method 1
                    lattice = self.parent.ui.method1_lattice_value_2.text()
                    if not is_float(lattice):
                        self.parent.ui.method1_lattice_error.setVisible(True)
                    else:
                        self.parent.ui.method1_lattice_error.setVisible(False)

                    # check validity of table
                    is_valid = self.is_table_valid(table_ui=self.parent.ui.method1_tableWidget)
                    if is_valid:
                        self.parent.ui.user_defined_method1_table_error.setVisible(False)
                    else:
                        self.parent.ui.user_defined_method1_table_error.setVisible(True)

                else:  # method 2
                    # check validity of table
                    is_valid = self.is_table_valid(table_ui=self.parent.ui.method2_tableWidget)
                    if is_valid:
                        self.parent.ui.user_defined_method2_table_error.setVisible(False)
                    else:
                        self.parent.ui.user_defined_method2_table_error.setVisible(True)

    def is_table_valid(self, table_ui=None):
        o_table = TableHandler(table_ui=table_ui)
        nbr_row = o_table.row_count()
        at_least_one_row_valid = False
        for _row in np.arange(nbr_row):
            h = o_table.get_item_str_from_cell(row=_row, column=0)
            k = o_table.get_item_str_from_cell(row=_row, column=1)
            l = o_table.get_item_str_from_cell(row=_row, column=2)  # noqa E741
            d0_or_lambda = o_table.get_item_str_from_cell(row=_row, column=3)

            if (
                ((h is None) or h == "")
                and ((k is None) or k == "")
                and ((l is None) or l == "")
                and ((d0_or_lambda is None) or d0_or_lambda == "")
            ):
                continue

            elif h is None:
                return False

            elif k is None:
                return False

            elif l is None:
                return False

            elif d0_or_lambda is None:
                return False

            if (h.strip() == "") and (k.strip() == "") and (l.strip() == "") and (d0_or_lambda.strip() == ""):
                continue

            if not is_int(h):
                return False

            if not is_int(k):
                return False

            if not is_int(l):
                return False

            if not is_float(d0_or_lambda):
                return False

            at_least_one_row_valid = True

        return at_least_one_row_valid


class MaterialPreDefined(Material):
    def update_pre_defined_widgets(self):
        element_selected = self.parent.ui.pre_defined_list_of_elements.currentText()
        _handler = BraggEdge(material=element_selected)
        _crystal_structure = _handler.metadata["crystal_structure"][element_selected]
        _lattice = str(_handler.metadata["lattice"][element_selected])

        o_calculator = BraggEdgeElementCalculator(
            element_name=element_selected,
            lattice_value=_lattice,
            crystal_structure=_crystal_structure,
        )
        o_calculator.run()

        selected_element_bragg_edges_array = o_calculator.lambda_array
        selected_element_hkl_array = o_calculator.hkl_array

        self.parent.ui.pre_defined_lattice_value.setText(_lattice)
        self.parent.ui.pre_defined_crystal_value.setText(_crystal_structure)

        o_table = TableHandler(table_ui=self.parent.ui.pre_defined_tableWidget)
        o_table.remove_all_rows()

        _row = 0
        for _hkl, _lambda in zip(selected_element_hkl_array, selected_element_bragg_edges_array):
            o_table.insert_empty_row(row=_row)
            _h = _hkl[0]
            _k = _hkl[1]
            _l = _hkl[2]

            o_table.insert_item(row=_row, column=0, editable=False, value=_h)

            o_table.insert_item(row=_row, column=1, editable=False, value=_k)

            o_table.insert_item(row=_row, column=2, editable=False, value=_l)

            o_table.insert_item(row=_row, column=3, editable=False, value=_lambda, format_str="{:.3f}")

            _row += 1


class MaterialUserDefined(Material):
    def user_defined_element_name_changed(self):
        element_name = self.parent.ui.user_defined_element_name.text()
        if element_name.strip() == "":
            display_error_message = True
        else:
            display_error_message = False
        self.parent.ui.user_defined_name_error.setVisible(display_error_message)


class MaterialUserDefinedMethod1(Material):
    def fill_fields_with_selected_element_clicked(self):
        element_selected = self.parent.ui.user_defined_list_of_elements.currentText()
        if element_selected == "None":
            # clear table
            o_table = TableHandler(table_ui=self.parent.ui.method1_tableWidget)
            o_table.remove_all_rows()
            self.parent.ui.method1_lattice_value_2.blockSignals(True)
            self.parent.ui.method1_crystal_value_2.blockSignals(True)

            self.parent.ui.method1_lattice_value_2.setText("")
            self.parent.ui.method1_crystal_value_2.setCurrentIndex(0)

            self.parent.ui.method1_lattice_value_2.blockSignals(False)
            self.parent.ui.method1_crystal_value_2.blockSignals(False)
            return

        _handler = BraggEdge(material=element_selected)
        _crystal_structure = _handler.metadata["crystal_structure"][element_selected]
        _lattice = str(_handler.metadata["lattice"][element_selected])

        o_calculator = BraggEdgeElementCalculator(
            element_name=element_selected,
            lattice_value=_lattice,
            crystal_structure=_crystal_structure,
        )
        o_calculator.run()

        selected_element_bragg_edges_array = o_calculator.lambda_array
        selected_element_hkl_array = o_calculator.hkl_array

        self.parent.ui.method1_lattice_value_2.blockSignals(True)
        self.parent.ui.method1_crystal_value_2.blockSignals(True)

        self.parent.ui.method1_lattice_value_2.setText(f"{float(_lattice):.3f}")
        _row_to_select = self.parent.ui.method1_crystal_value_2.findText(_crystal_structure)
        self.parent.ui.method1_crystal_value_2.setCurrentIndex(_row_to_select)

        self.parent.ui.method1_lattice_value_2.blockSignals(False)
        self.parent.ui.method1_crystal_value_2.blockSignals(False)

        self.fill_table_method1(
            selected_element_hkl_array=selected_element_hkl_array,
            selected_element_bragg_edges_array=selected_element_bragg_edges_array,
        )

    def user_defined_element_name_changed(self):
        element_name = self.parent.ui.user_defined_element_name.text()
        lattice_value = self.parent.ui.method1_lattice_value_2.text()
        if not is_float(lattice_value):
            o_table = TableHandler(table_ui=self.parent.ui.method1_tableWidget)
            o_table.remove_all_rows()
            return

        crystal_structure = self.parent.ui.method1_crystal_value_2.currentText()

        o_calculator = BraggEdgeElementCalculator(
            element_name=element_name,
            lattice_value=lattice_value,
            crystal_structure=crystal_structure,
        )
        o_calculator.run()

        selected_element_bragg_edges_array = o_calculator.lambda_array
        selected_element_hkl_array = o_calculator.hkl_array

        self.fill_table_method1(
            selected_element_hkl_array=selected_element_hkl_array,
            selected_element_bragg_edges_array=selected_element_bragg_edges_array,
        )

    def fill_table_method1(self, selected_element_hkl_array=None, selected_element_bragg_edges_array=None):
        o_table = TableHandler(table_ui=self.parent.ui.method1_tableWidget)
        o_table.remove_all_rows()

        self.parent.ui.method1_tableWidget.blockSignals(True)
        _row = 0
        for _hkl, _lambda in zip(selected_element_hkl_array, selected_element_bragg_edges_array):
            o_table.insert_empty_row(row=_row)
            _h = _hkl[0]
            _k = _hkl[1]
            _l = _hkl[2]

            o_table.insert_item(row=_row, column=0, value=_h)

            o_table.insert_item(row=_row, column=1, value=_k)

            o_table.insert_item(row=_row, column=2, value=_l)

            o_table.insert_item(row=_row, column=3, value=_lambda, format_str="{:.3f}")

            _row += 1

        self.parent.ui.method1_tableWidget.blockSignals(False)

    def lattice_crystal_changed(self):
        """recalculate the hkl and lambda 0 of the table"""
        self.user_defined_element_name_changed()


class MaterialUserDefinedMethod2(Material):
    def clear_user_defined_method2_table(self):
        o_table = TableHandler(table_ui=self.parent.ui.method2_tableWidget)
        o_table.remove_all_rows()
        for _row in np.arange(10):
            o_table.insert_empty_row(row=_row)



#!/usr/bin/env python
"""
Initialization module
"""

import pyqtgraph as pg
from pyqtgraph.dockarea import Dock, DockArea
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QHBoxLayout, QProgressBar, QRadioButton, QSizePolicy, QSpacerItem, QVBoxLayout, QWidget

from ibeatles import interact_me_style
from ibeatles.tools.tof_combine import ANGSTROMS, LAMBDA, MICRO, settings_image
from ibeatles.tools.tof_combine.utilities.table_handler import TableHandler


class Initialization:
    def __init__(self, parent=None):
        self.parent = parent

    def all(self):
        self.pyqtgraph_combine()
        self.statusbar()
        self.splitter()
        self.table()
        self.tab()
        self.labels()
        self.widgets()

    def tab(self):
        self.parent.ui.combine_bottom_tabWidget.setTabIcon(2, QIcon(settings_image))

    def statusbar(self):
        self.parent.eventProgress = QProgressBar(self.parent.ui.statusbar)
        self.parent.eventProgress.setMinimumSize(20, 14)
        self.parent.eventProgress.setMaximumSize(540, 100)
        self.parent.eventProgress.setVisible(False)
        self.parent.ui.statusbar.addPermanentWidget(self.parent.eventProgress)
        self.parent.setStyleSheet("QStatusBar{padding-left:8px;color:red;font-weight:bold;}")

    def splitter(self):
        self.parent.ui.combine_horizontal_splitter.setSizes([100, 0])
        self.parent.ui.combine_horizontal_splitter.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/horizontal_splitter_handle.png");
                                     }
                                     """)
        self.parent.ui.combine_horizontal_splitter.setHandleWidth(15)

    def table(self):
        # combine table
        o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        column_sizes = [100, 100, 500]
        o_table.set_column_sizes(column_sizes=column_sizes)

    def labels(self):
        # combine tab
        self.parent.ui.combine_detector_offset_units.setText(MICRO + "s")

    def pyqtgraph_combine(self):
        area = DockArea()
        self.parent.ui.area = area
        d1 = Dock("Image Preview", size=(200, 300))
        d2 = Dock("ROI profile", size=(200, 100))

        area.addDock(d1, "top")
        area.addDock(d2, "bottom")

        # preview - top widget
        image_view = pg.ImageView(view=pg.PlotItem())
        image_view.ui.roiBtn.hide()
        image_view.ui.menuBtn.hide()
        self.parent.combine_image_view = image_view
        image_view.scene.sigMouseMoved.connect(self.parent.mouse_moved_in_combine_image_preview)
        d1.addWidget(image_view)

        # plot and x-axis radio buttons - bottom widgets
        profile = pg.PlotWidget(title="")
        profile.plot()
        self.parent.combine_profile_view = profile
        bottom_layout = QVBoxLayout()
        bottom_layout.addWidget(profile)
        # xaxis radio buttons
        spacer_left = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        file_index_radio_button = QRadioButton("File Index")
        file_index_radio_button.setChecked(True)
        self.parent.combine_file_index_radio_button = file_index_radio_button
        self.parent.combine_file_index_radio_button.clicked.connect(self.parent.combine_xaxis_changed)
        tof_radio_button = QRadioButton("TOF (" + MICRO + "s)")
        self.parent.tof_radio_button = tof_radio_button
        self.parent.tof_radio_button.clicked.connect(self.parent.combine_xaxis_changed)
        lambda_radio_button = QRadioButton(LAMBDA + " (" + ANGSTROMS + ")")
        self.parent.lambda_radio_button = lambda_radio_button
        self.parent.lambda_radio_button.clicked.connect(self.parent.combine_xaxis_changed)
        spacer_right = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        axis_layout = QHBoxLayout()
        axis_layout.addItem(spacer_left)
        axis_layout.addWidget(file_index_radio_button)
        axis_layout.addWidget(tof_radio_button)
        axis_layout.addWidget(lambda_radio_button)
        axis_layout.addItem(spacer_right)
        bottom_widget = QWidget()
        bottom_widget.setLayout(axis_layout)
        bottom_layout.addWidget(bottom_widget)
        widget = QWidget()
        widget.setLayout(bottom_layout)
        d2.addWidget(widget)

        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(area)
        self.parent.ui.combine_widget.setLayout(vertical_layout)
        self.parent.ui.combine_widget.setEnabled(False)

    def widgets(self):
        self.parent.ui.combine_refresh_top_folder_pushButton.setEnabled(False)
        self.parent.ui.combine_select_top_folder_pushButton.setStyleSheet(interact_me_style)



#!/usr/bin/env python
"""
This module provides a class to load files.
"""

import glob
import os

import numpy as np
from qtpy.QtWidgets import QApplication

from ibeatles import DataType
from ibeatles.utilities.file_handler import FileHandler
from ibeatles.utilities.image_handler import ImageHandler


class LoadFiles:
    # class variables
    image_array = []
    list_of_files = []
    data = []

    def __init__(self, parent=None, image_ext=".tif", folder=None, list_of_files=None):
        self.parent = parent
        self.image_ext = image_ext
        self.folder = folder

        self.retrieve_list_of_files(list_of_files=list_of_files)
        self.retrieve_data()

    def retrieve_list_of_files(self, list_of_files=None):
        _folder = self.folder
        _image_ext = self.image_ext

        if list_of_files is None:
            _list_of_files = glob.glob(_folder + "/*" + _image_ext)
        else:
            _list_of_files = list_of_files

        self.list_of_files_full_name = _list_of_files
        short_list_of_files = []
        self.folder = os.path.dirname(_list_of_files[0]) + "/"
        for _file in _list_of_files:
            _short_file = os.path.basename(_file)
            short_list_of_files.append(_short_file)

        short_list_of_files = FileHandler.cleanup_list_of_files(short_list_of_files)
        self.list_of_files = short_list_of_files

    def retrieve_data(self):
        self.image_array = []

        self.parent.eventProgress.setMinimum(0)
        self.parent.eventProgress.setMaximum(len(self.list_of_files))
        self.parent.eventProgress.setValue(0)
        self.parent.eventProgress.setVisible(True)

        for _index, _file in enumerate(self.list_of_files):
            full_file_name = os.path.join(self.folder, _file)
            o_handler = ImageHandler(parent=self.parent, filename=full_file_name)
            _data = o_handler.get_data()
            self.image_array.append(_data)
            self.parent.eventProgress.setValue(_index + 1)
            QApplication.processEvents()

            if _index == 0:
                [height, width] = np.shape(_data)
                self.parent.data_metadata[DataType.normalized]["size"] = {
                    "width": width,
                    "height": height,
                }

        self.parent.eventProgress.setVisible(False)

    @classmethod
    def load_interactive_data(cls, parent=None, list_tif_files=None):
        dict = {"width": None, "height": None, "image_array": []}

        parent.eventProgress.setMinimum(0)
        parent.eventProgress.setMaximum(len(list_tif_files))
        parent.eventProgress.setValue(0)
        parent.eventProgress.setVisible(True)

        for _index, _file in enumerate(list_tif_files):
            o_handler = ImageHandler(parent=parent, filename=_file)
            _data = o_handler.get_data()
            dict["image_array"].append(_data)
            parent.eventProgress.setValue(_index + 1)
            QApplication.processEvents()

            if _index == 0:
                [height, width] = np.shape(_data)
                dict["width"] = width
                dict["height"] = height

        parent.eventProgress.setVisible(False)

        return dict
